from __future__ import annotations

import asyncio
import typing

import toolstr
import numpy as np

from ctc import config
from ctc import evm
from ctc import spec
from . import aave_spec
from . import aave_interest_rates
from . import aave_oracle


async def async_print_aave_addresses(
    *,
    verbose: bool = False,
    block: spec.BlockNumberReference | None = None,
    max_width: int | None = None,
    context: spec.Context = None,
) -> None:

    from ctc import cli

    indent = 4

    styles = cli.get_cli_styles()

    toolstr.print_text_box('Aave V2 Addresses', style=styles['title'])

    network = config.get_context_chain_id(context)
    toolstr.print('network = ' + str(network), style=styles['comment'])

    # print general contracts
    contracts = [
        'PriceOracle',
        'LendingPoolProvider',
        'LendingPool',
        'IncentivesController',
        'Collector',
    ]
    rows = []
    for contract in contracts:
        address = aave_spec.get_aave_address(contract, context=context)
        row = [contract, address]
        rows.append(row)
    print()
    if max_width is not None:
        max_table_width = max_width - indent
    else:
        max_table_width = None
    toolstr.print_table(
        rows,
        indent=indent,
        border=styles['comment'],
        column_styles=[styles['option'], styles['metavar']],
        max_table_width=max_table_width,
    )

    # print token markets
    token_markets = await aave_interest_rates.async_get_token_markets(
        block=block, context=context
    )

    print()
    print()
    toolstr.print_header('Token Markets', style=styles['title'])
    print()
    if verbose:

        rows = []
        for token_market in token_markets:
            address_labels = [
                'underlying',
                'atoken',
                'variable debt',
                'stable debt',
                'interest rate',
            ]
            addresses = [
                token_market['underlying'],
                token_market['reserve_data']['atoken_address'],
                token_market['reserve_data']['variable_debt_token_address'],
                token_market['reserve_data']['stable_debt_token_address'],
                token_market['reserve_data']['interest_rate_strategy_address'],
            ]
            address_pieces = [
                toolstr.add_style(label, styles['comment'])
                + ' '
                + toolstr.add_style(address, styles['metavar'])
                for label, address in zip(address_labels, addresses)
            ]

            row = [token_market['symbol'], '\n'.join(address_pieces)]
            rows.append(row)
        labels = [
            'symbol',
            'addresses',
        ]

        rows = sorted(rows, key=lambda row: row[0])

        toolstr.print_multiline_table(
            rows,
            indent=indent,
            labels=labels,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'symbol': styles['option'],
            },
            max_table_width=max_table_width,
        )

    else:
        rows = []
        for token_market in token_markets:
            row = [
                token_market['symbol'],
                token_market['reserve_data']['atoken_address'],
            ]
            rows.append(row)
        labels = [
            'symbol',
            'atoken',
        ]

        rows = sorted(rows, key=lambda row: row[0])

        toolstr.print_table(
            rows,
            indent=indent,
            labels=labels,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles={
                'symbol': styles['option'],
                'atoken': styles['metavar'],
            },
            max_table_width=max_table_width,
        )


async def async_print_token_markets_summary(
    *,
    verbose: bool = False,
    block: spec.BlockNumberReference | None = None,
    max_width: int | None = None,
    context: spec.Context = None,
) -> None:

    from ctc import cli

    if block is None:
        block = 'latest'

    token_markets = await aave_interest_rates.async_get_token_markets(
        block=block, context=context
    )

    total_supplies_coroutine = evm.async_get_erc20s_total_supplies(
        [
            token_market['reserve_data']['atoken_address']
            for token_market in token_markets
        ],
        block=block,
        context=context,
    )
    total_supplies_task = asyncio.create_task(total_supplies_coroutine)

    coroutines = [
        aave_interest_rates.async_get_interest_rates(
            reserve_data=token_market['reserve_data'],
            block=block,
            context=context,
        )
        for token_market in token_markets
    ]
    interest_rates_task = asyncio.gather(*coroutines)

    reserves_list = [
        token_market['underlying'] for token_market in token_markets
    ]
    prices_coroutine = aave_oracle.async_get_asset_prices(
        reserves_list, block=block, context=context
    )
    prices_task = asyncio.create_task(prices_coroutine)

    reserve_balances_coroutines = [
        evm.async_get_erc20_balance(
            wallet=token_market['reserve_data']['atoken_address'],
            token=token_market['underlying'],
            block=block,
            context=context,
        )
        for token_market in token_markets
    ]
    reserve_balances = await asyncio.gather(*reserve_balances_coroutines)

    interest_rates = await interest_rates_task
    prices = await prices_task
    total_supplies = await total_supplies_task

    rows = []
    tvls = []
    tvbs = []
    for r in range(len(token_markets)):
        supply_apy = interest_rates[r]['supply_apy']
        borrow_apy = interest_rates[r]['borrow_apy']
        total_supply = total_supplies[r]
        tvl = total_supply * prices[r]
        tvls.append(tvl)

        liquidity = reserve_balances[r]
        tvb = tvl - liquidity * prices[r]
        tvbs.append(tvb)

        row = [
            token_markets[r]['symbol'],
            supply_apy,
            borrow_apy,
            tvl,
            tvb,
        ]
        if verbose:
            row.append(tvb / tvl)
            row.append(prices[r])
        rows.append(row)

    labels = [
        'token',
        'supply APY',
        'borrow APY',
        'TVL',
        'TVB',
    ]
    if verbose:
        labels.append('util %')
        labels.append('price')

    # sort rows
    sort_by = [
        #     'supply APY',
        'TVL',
    ]
    sort_indices = [labels.index(column) for column in sort_by]
    ascending = False
    rows = sorted(
        rows,
        reverse=(not ascending),
        key=lambda row: tuple(row[index] for index in sort_indices),
    )

    # special formatting
    oom_columns = [
        'supply APY',
        'borrow APY',
    ]
    for row in rows:
        for column in oom_columns:
            index = labels.index(column)
            item = row[index]
            if item > 100:
                item = toolstr.format(item * 100, order_of_magnitude=True) + '%'
                row[index] = item

    rows.append([''] * len(labels))
    total_row = [
        'TOTAL',
        '',
        '',
        toolstr.format(
            sum(tvls),
            order_of_magnitude=True,
            decimals=2,
            trailing_zeros=True,
            prefix='$',
        ),
        toolstr.format(
            sum(tvbs),
            order_of_magnitude=True,
            decimals=2,
            trailing_zeros=True,
            prefix='$',
        ),
    ]
    if verbose:
        total_row.extend(['', ''])
    total_row = [toolstr.add_style(cell, 'bold') for cell in total_row]
    rows.append(total_row)

    styles = cli.get_cli_styles()
    toolstr.print_text_box('Aave V2 Markets', style=styles['title'])
    print()
    toolstr.print_table(
        rows,
        labels=labels,
        column_formats={
            'supply APY': {
                'percentage': True,
                'trailing_zeros': True,
                'decimals': 3,
            },
            'borrow APY': {
                'percentage': True,
                'trailing_zeros': True,
                'decimals': 3,
            },
            'TVL': {
                'order_of_magnitude': True,
                'trailing_zeros': True,
                'prefix': '$',
                'decimals': 2,
            },
            'TVB': {
                'order_of_magnitude': True,
                'trailing_zeros': True,
                'prefix': '$',
                'decimals': 2,
            },
            'util %': {
                'percentage': True,
                'trailing_zeros': True,
                'decimals': 2,
            },
            'price': {
                'prefix': '$',
                'trailing_zeros': True,
                'decimals': 2,
            },
        },
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'token': styles['metavar'],
            'supply APY': styles['description'],
            'borrow APY': styles['description'],
            'TVL': styles['description'],
            'TVB': styles['description'],
            'util %': styles['description'],
            'price': styles['description'],
        },
        max_table_width=max_width,
    )
    print()
    toolstr.print('network = ethereum', style=styles['comment'])


async def async_get_token_market_summary(
    token: str | spec.Address,
    *,
    blocks: typing.Sequence[int],
    verbose: bool = False,
    context: spec.Context = None,
) -> typing.Mapping[str, typing.Any]:

    if not evm.is_address_str(token):
        token_address = await evm.async_get_erc20_address(
            token, context=context
        )
    else:
        token_address = token

    reserve_data_by_block = (
        await aave_interest_rates.async_get_reserve_data_by_block(
            asset=token_address,
            blocks=blocks,
            context=config.update_context(
                context, merge_provider={'chunk_size': 1}
            ),
        )
    )

    atoken_address = reserve_data_by_block['atoken_address'][0]

    interest_rates_task = asyncio.create_task(
        aave_interest_rates.async_get_interest_rates_by_block(
            token=token_address,
            blocks=blocks,
            reserve_data_by_block=reserve_data_by_block,
        )
    )

    # compute tvls
    total_supplies_task = asyncio.create_task(
        evm.async_get_erc20_total_supply_by_block(
            atoken_address,
            blocks=blocks,
            context=context,
        )
    )

    # compute tvbs
    balances_task = asyncio.create_task(
        evm.async_get_erc20_balance_by_block(
            wallet=atoken_address,
            token=token,
            blocks=blocks,
            context=context,
        )
    )

    total_supplies = await total_supplies_task
    total_supplies_array = np.array(total_supplies)
    prices_task = asyncio.create_task(
        aave_oracle.async_get_asset_price_by_block(
            token_address,
            blocks=blocks,
            context=context,
        )
    )

    prices = await prices_task
    prices_array = np.array(prices)
    tvls = total_supplies_array * prices_array

    balances = await balances_task
    balances_array = np.array(balances)
    tvbs = (total_supplies_array - balances_array) * prices_array

    if verbose:
        utilization = tvbs / tvls
    else:
        utilization = None

    # compute interest rates
    interest_rates = await interest_rates_task
    supply_apy = np.array(interest_rates['supply_apy']) * 100
    borrow_apy = np.array(interest_rates['borrow_apy']) * 100

    return {
        'prices': prices_array,
        'tvl': tvls,
        'tvb': tvbs,
        'supply_apy': supply_apy,
        'borrow_apy': borrow_apy,
        'utilization': utilization,
    }


async def async_print_token_market_summary(
    token: str | spec.Address,
    *,
    blocks: typing.Sequence[int],
    verbose: bool = False,
    context: spec.Context = None,
) -> None:

    from ctc import cli

    summary = await async_get_token_market_summary(
        token=token, blocks=blocks, verbose=verbose, context=context
    )
    prices_array = summary['prices']
    tvls = summary['tvl']
    tvbs = summary['tvb']
    supply_apy = summary['supply_apy']
    borrow_apy = summary['borrow_apy']
    utilization = summary['utilization']

    timestamps_task = asyncio.create_task(
        evm.async_get_block_timestamps(blocks, context=context)
    )

    styles = cli.get_cli_styles()
    toolstr.print_text_box(
        'Summary of ' + token + ' on Aave v2', style=styles['title']
    )

    percentage_kwargs = {
        'yaxis_kwargs': {
            'tick_label_format': {
                'postfix': '%',
                'decimals': 2,
                'trailing_zeros': True,
            }
        }
    }
    dollar_kwargs = {'yaxis_kwargs': {'tick_label_format': {'prefix': '$'}}}
    plots = [
        ('Supply APY', supply_apy, percentage_kwargs),
        ('Borrow APY', borrow_apy, percentage_kwargs),
        ('TVL', tvls, dollar_kwargs),
        ('TVB', tvbs, dollar_kwargs),
    ]
    if verbose:
        plots.append(('Utilization', utilization * 100, percentage_kwargs))
        plots.append(('Price', prices_array, dollar_kwargs))
    timestamps = await timestamps_task
    for title, data, plot_kwargs in plots:
        print()
        plot = toolstr.render_line_plot(
            xvals=timestamps,
            yvals=data,
            n_rows=10,
            n_columns=60,
            line_style=styles['description'],
            chrome_style=styles['comment'],
            tick_label_style=styles['metavar'],
            char_dict=config.get_cli_chart_charset(),
            **plot_kwargs,  # type: ignore
        )
        toolstr.print(
            toolstr.hjustify(title, 'center', 70),
            indent=4,
            style=styles['title'],
        )
        print()
        toolstr.print(plot, indent=4)
        print()
        print()




from __future__ import annotations

import typing

from ctc import evm
from ctc import rpc
from ctc import spec

from . import curve_spec


async def async_get_pool_tokens(
    pool: spec.Address,
    *,
    n_tokens: int | None = None,
    context: spec.Context = None,
) -> list[spec.Address]:
    import asyncio

    old_pools = {
        '0x79a8c46dea5ada233abaffd40f3a0a2b1e5a4f27',
        '0xa2b47e3d5c44877cca798226b7b8118f9bfb7a56',
        '0x06364f10b501e868329afbc005b3492902d6c763',
        '0x93054188d876f558f4a66b2ef1d97d16edf0895b',
        '0x7fc77b5c7614e1533320ea6ddc2eb61fa00a9714',
        '0xa5407eae9ba41422680e2e00537571bcc53efbfd',
        '0x52ea46506b9cc5ef470c5bf89f17dc28bb35d85c',
        '0x45f783cce6b7ff23b2ab2d70e416cdb7d6055f51',
    }
    if pool in old_pools:
        function_abi: spec.FunctionABI = {
            'name': 'coins',
            'inputs': [{'type': 'int128'}],
            'outputs': [{'type': 'address'}],
        }
    else:
        function_abi = {
            'name': 'coins',
            'inputs': [{'type': 'uint256'}],
            'outputs': [{'type': 'address'}],
        }

    if n_tokens is None:
        token_addresses = []
        t = 0
        while True:
            try:
                token_address = await rpc.async_eth_call(
                    to_address=pool,
                    function_abi=function_abi,
                    function_parameters=[t],
                    context=context,
                )
                token_addresses.append(token_address)
                t += 1
            except spec.RpcException:
                break
    else:
        address_coroutines = [
            rpc.async_eth_call(
                to_address=pool,
                function_abi=function_abi,
                function_parameters=[i],
                context=context,
            )
            for i in range(n_tokens)
        ]
        token_addresses = await asyncio.gather(*address_coroutines)

    return token_addresses


async def async_get_token_index(
    token: typing.Union[int, spec.Address, str],
    pool: spec.Address | None = None,
    *,
    metadata: curve_spec.CurvePoolMetadata | None = None,
    n_tokens: int | None = None,
    context: spec.Context = None,
) -> int:

    if isinstance(token, int):
        # token already an index
        return token

    elif isinstance(token, str) and token.startswith('0x'):
        # token an address
        if metadata is not None:
            pool_addresses = metadata['token_addresses']
        elif pool is not None:
            pool_addresses = await async_get_pool_tokens(
                pool=pool,
                n_tokens=n_tokens,
                context=context,
            )
        else:
            raise Exception('must specify more parameters')
        return pool_addresses.index(token)

    elif isinstance(token, str):
        # token a symbol
        if metadata is not None:
            pool_addresses = metadata['token_addresses']
        elif pool is not None:
            pool_addresses = await async_get_pool_tokens(
                pool=pool,
                n_tokens=n_tokens,
                context=context,
            )
        else:
            raise Exception('must specify more parameters')
        token_symbols = await evm.async_get_erc20s_symbols(
            pool_addresses,
            context=context,
        )
        return token_symbols.index(token)

    else:
        raise Exception()


async def async_get_pool_metadata(
    pool: spec.Address,
    *,
    n_tokens: int | None = None,
    context: spec.Context = None,
) -> curve_spec.CurvePoolMetadata:

    import asyncio

    function_abi: spec.FunctionABI = {
        'inputs': [],
        'name': 'A',
        'outputs': [
            {
                'name': '',
                'type': 'uint256',
            },
        ],
        'stateMutability': 'view',
        'type': 'function',
    }

    a_coroutine = rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        function_parameters=[],
        context=context,
    )
    a_task = asyncio.create_task(a_coroutine)

    # get addresses
    token_addresses = await async_get_pool_tokens(
        pool,
        n_tokens=n_tokens,
        context=context,
    )

    # get additional metadata
    symbol_coroutine = evm.async_get_erc20s_symbols(
        token_addresses,
        context=context,
    )
    decimal_coroutine = evm.async_get_erc20s_decimals(
        token_addresses,
        context=context,
    )
    symbol_task = asyncio.create_task(symbol_coroutine)
    decimal_task = asyncio.create_task(decimal_coroutine)

    # await results
    token_symbols = await symbol_task
    token_decimals = await decimal_task
    A = await a_task

    return {
        'token_addresses': token_addresses,
        'token_symbols': token_symbols,
        'token_decimals': token_decimals,
        'A': A,
        # fee:
    }




from __future__ import annotations

import toolcli
import toolstr

from ctc import config
from ctc import evm
from ctc import cli
from ctc.protocols.chainlink_utils import chainlink_db


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_chainlink_ls_command,
        'help': 'list all Chainlink feeds',
        'args': [
            {
                'name': 'query',
                'nargs': '?',
                'help': 'partial match of feed name, case insensitive',
            },
            {
                'name': '--network',
                'help': 'network to list feeds of',
            },
        ],
        'examples': [
            '',
            'DAI',
        ],
    }


async def async_chainlink_ls_command(
    query: str, network: str | int | None
) -> None:

    context = config.create_user_input_context(network=network)

    oracle_feeds = await chainlink_db.async_query_feeds(context=context)
    if oracle_feeds is None:
        return
    rows = []
    for feed in oracle_feeds:
        if feed is None:
            continue
        if query is not None and query.lower() not in feed['name'].lower():
            continue
        row = [
            feed['name'],
            feed['deviation'],
            feed['heartbeat'],
            feed['address'][:42],
        ]
        rows.append(row)
    labels = ['name', 'delta', 'rate', 'address']
    styles = cli.get_cli_styles()
    network_name = evm.get_network_name(config.get_context_chain_id(context))
    toolstr.print_text_box(
        'Chainlink feeds on ' + network_name.title(),
        style=styles['title'],
    )
    toolstr.print_table(
        rows,
        labels=labels,
        column_gap=1,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'name': styles['option'],
            'delta': styles['description'],
            'rate': styles['description'],
            'address': styles['metavar'],
        },
        max_column_widths={'name': 14},
    )




"""Track legacy type specification of Config starting from version 0.2.1

These definitions should be mostly self-contained
- do not want them altered by upstream changes to definition
- thus, provide definitions for:
    - Provider
    - NetworkMetadata
    - toolsql.DBConfig
"""

from __future__ import annotations

import typing

from ctc import spec

if typing.TYPE_CHECKING:
    from typing_extensions import Literal
    from typing_extensions import TypedDict

    class LegacyConfig__3_0_0(TypedDict):
        config_spec_version: str
        data_dir: str
        providers: typing.Mapping[str, LegacyProvider__0_3_0]
        networks: typing.Mapping[spec.ChainId, LegacyNetworkMetadata__0_3_0]
        default_network: spec.ChainId | None
        default_providers: typing.Mapping[spec.ChainId, str]
        db_configs: typing.Mapping[str, ToolsqlDBConfig]
        log_rpc_calls: bool
        log_sql_queries: bool

    class LegacyConfig__0_2_3(TypedDict):
        config_spec_version: str
        data_dir: str
        providers: typing.Mapping[str, LegacyProvider__0_2_1]
        networks: typing.Mapping[spec.NetworkName, LegacyNetworkMetadata__0_2_1]
        network_defaults: LegacyConfigNetworkDefaults

    class LegacyConfig__0_2_1(TypedDict):
        version: str
        data_dir: str
        providers: typing.Mapping[str, LegacyProvider__0_2_1]
        networks: typing.Mapping[spec.NetworkName, LegacyNetworkMetadata__0_2_1]
        network_defaults: LegacyConfigNetworkDefaults

    class LegacyConfigNetworkDefaults(TypedDict):
        default_network: spec.NetworkName
        default_providers: typing.Mapping[spec.NetworkName, str]

    #
    # # network metadata
    #

    class LegacyNetworkMetadata__0_3_0(TypedDict):
        name: str | None
        chain_id: int
        block_explorer: str | None

    class LegacyNetworkMetadata__0_2_1(TypedDict):
        name: str
        chain_id: int
        block_explorer: str

    #
    # # provider
    #

    class LegacyProvider__0_3_0(TypedDict, total=True):
        url: str
        name: typing.Optional[str]
        network: typing.Optional[str | int]
        protocol: Literal['http', 'wss', 'ipc']
        session_kwargs: typing.Optional[typing.Mapping[typing.Any, typing.Any]]
        chunk_size: typing.Optional[int]

    class LegacyProvider__0_2_1(TypedDict, total=True):
        url: str
        name: typing.Optional[str]
        network: typing.Optional[str]
        protocol: Literal['http', 'wss', 'ipc']
        session_kwargs: typing.Optional[typing.Mapping[typing.Any, typing.Any]]
        chunk_size: typing.Optional[int]

    #
    # # toolsql
    #

    class ToolsqlDBConfig(TypedDict, total=False):
        dbms: Literal['sqlite', 'postgresql']
        path: str
        engine: str
        hostname: str
        port: int
        database: str
        username: str
        password: str
        socket: str
        socket_dir: str
        timeout: typing.Union[int, float]
        pool_timeout: typing.Union[int, float]




from __future__ import annotations

from ctc import rpc
from ctc import spec

from .. import uniswap_v3_spec


async def async_pool_observe(
    seconds_agos: list[int],
    pool: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> dict[str, int]:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'observe', 'pool'
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        function_parameters=[seconds_agos],
        context=context,
        block_number=block,
    )
    return {
        'tick_cumulatives': result[0],
        'seconds_per_liquidity_cumulative_x128': result[1],
    }


async def async_pool_snapshot_cumulatives_inside(
    *,
    tick_lower: int,
    tick_upper: int,
    pool: spec.Address,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> dict[str, int]:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'snapshotCumulativesInside',
        'pool',
    )
    result = await rpc.async_eth_call(
        to_address=pool,
        function_abi=function_abi,
        function_parameters=[tick_lower, tick_upper],
        context=context,
        block_number=block,
    )
    return {
        'tick_cumulative_inside': result[0],
        'seconds_per_liquidity_x128': result[1],
        'seconds_inside': result[2],
    }



from __future__ import annotations

import typing

import ctc
from ctc import evm
from ctc import spec

from .. import rpc_request


def construct_eth_block_number() -> spec.RpcSingularRequest:
    return rpc_request.create('eth_blockNumber', [])


def construct_eth_get_block_by_hash(
    block_hash: spec.BinaryData,
    *,
    include_full_transactions: bool = False,
) -> spec.RpcSingularRequest:
    encoded_block_hash = evm.to_hex(block_hash)
    parameters = [encoded_block_hash, include_full_transactions]
    return rpc_request.create('eth_getBlockByHash', parameters)


def construct_eth_get_block_by_number(
    block_number: spec.BlockNumberReference,
    *,
    include_full_transactions: bool = False,
) -> spec.RpcSingularRequest:
    encoded_block_number = evm.encode_block_number(block_number)

    parameters = [encoded_block_number, include_full_transactions]
    return rpc_request.create(
        method='eth_getBlockByNumber',
        parameters=parameters,
    )


def construct_eth_get_uncle_count_by_block_hash(
    block_hash: spec.BinaryData,
) -> spec.RpcSingularRequest:
    encoded_block_hash = evm.to_hex(block_hash)
    return rpc_request.create(
        method='eth_getUncleCountByBlockHash',
        parameters=[encoded_block_hash],
    )


def construct_eth_get_uncle_count_by_block_number(
    block_number: spec.BlockNumberReference,
) -> spec.RpcSingularRequest:
    encoded_block_number = evm.encode_block_number(block_number)
    return rpc_request.create(
        method='eth_getUncleCountByBlockNumber',
        parameters=[encoded_block_number],
    )


def construct_eth_get_uncle_by_block_hash_and_index(
    block_hash: spec.BinaryData, uncle_index: spec.BinaryData
) -> spec.RpcSingularRequest:
    encoded_block_hash = evm.to_hex(block_hash)
    encoded_uncle_index = evm.to_hex(uncle_index, keep_leading_0=False)

    return rpc_request.create(
        method='eth_getUncleByBlockHashAndIndex',
        parameters=[encoded_block_hash, encoded_uncle_index],
    )


def construct_eth_get_uncle_by_block_number_and_index(
    block_number: spec.BlockNumberReference, uncle_index: spec.BinaryData
) -> spec.RpcSingularRequest:
    encoded_block_number = evm.encode_block_number(block_number)
    encoded_uncle_index = evm.to_hex(uncle_index, keep_leading_0=False)

    return rpc_request.create(
        method='eth_getUncleByBlockNumberAndIndex',
        parameters=[encoded_block_number, encoded_uncle_index],
    )


def construct_eth_fee_history(
    block_number: spec.BlockNumberReference = 'latest',
    block_count: int = 1024,
    *,
    reward_percentiles: typing.Sequence[float] | None = None,
) -> spec.RpcSingularRequest:
    encoded_block_number = evm.encode_block_number(block_number)
    if reward_percentiles is None:
        reward_percentiles = []
    return rpc_request.create(
        method='eth_feeHistory',
        parameters=[
            ctc.to_hex(block_count, keep_leading_0=False),
            encoded_block_number,
            reward_percentiles,
        ],
    )




"""
## Resources
- https://github.com/TokenEngineeringCommunity/BalancerPools_Model
- https://token-engineering-balancer.gitbook.io/balancer-simulations/additional-code-and-instructions/balancer-the-python-edition/balancer_math.py
"""

from __future__ import annotations

from ctc import spec


def calc_out_given_in(
    *,
    token_amount_in: spec.Number,
    token_balance_in: spec.Number,
    token_weight_in: float,
    token_balance_out: spec.Number,
    token_weight_out: float,
    swap_fee: float,
) -> spec.Number:
    """
    adapted from https://github.com/TokenEngineeringCommunity/BalancerPools_Model
    """
    weight_ratio = token_weight_in / token_weight_out
    adjusted_in = token_amount_in * (1 - swap_fee)
    y = token_balance_in / (token_balance_in + adjusted_in)
    foo = pow(y, weight_ratio)
    bar = 1 - foo
    token_amount_out = token_balance_out * bar
    if not spec.is_number(token_amount_out):
        raise Exception('cannot compute result')
    return token_amount_out



from __future__ import annotations

import typing

import toolcli
import toolsql

from ctc import config
from ctc import spec
from ctc.spec.typedefs import db_types
from .. import schema_utils
from . import version_utils


def create_missing_tables(
    networks: typing.Sequence[spec.NetworkReference] | None = None,
    schema_names: typing.Sequence[str] | None = None,
    *,
    conn: toolsql.Connection,
    verbose: bool = True,
    confirm: bool = False,
) -> None:

    # get netowrks and schemas
    network_schema_names = schema_utils.get_network_schema_names()
    generic_schema_names = schema_utils.get_generic_schema_names()
    if networks is None:
        networks = config.get_networks_that_have_providers()
    if schema_names is None:
        schema_names = network_schema_names + generic_schema_names

    # get preamble
    if verbose or not confirm:
        n_schemas = len(schema_names)
        n_networks = len(networks)
        print('Using', n_schemas, 'schema(s) over', n_networks, 'network(s)')
        if len(schema_names) > 1:
            print('    - schemas:')
            for schema_name in schema_names:
                print('        -', schema_name)
        elif len(schema_names) == 1:
            print('    - schema:', schema_names[0])
        else:
            print('    - schema: [none specified]')
        if len(networks) > 1:
            print('    - networks:')
            for network in networks:
                print('        -', network)
        elif len(networks) == 1:
            print('    - network:', networks[0])
        else:
            print()
            print('No networks specified, creating no tables')

    # get missing tables
    schemas_to_create: list[
        tuple[spec.NetworkReference | None, spec.SchemaName]
    ] = []
    for schema_name_raw in schema_names:

        if schema_name_raw in schema_utils.get_all_schema_names():
            schema_name = typing.cast(db_types.SchemaName, schema_name_raw)
        else:
            raise Exception('unknown schema name: ' + str(schema_name_raw))

        # determine schema networks
        if schema_name in network_schema_names:
            schema_networks: typing.Sequence[
                spec.NetworkReference | None
            ] = networks
        else:
            schema_networks = [None]

        for schema_network in schema_networks:

            if schema_network is not None:
                schema = schema_utils.get_prepared_schema(
                    schema_name=schema_name,
                    context=dict(network=schema_network),
                )
            else:
                schema = schema_utils.get_raw_schema(schema_name=schema_name)

            # if any tables missing, mark schema for creation
            tables_in_db = toolsql.get_table_names(conn=conn)
            missing_tables = [
                table
                for table in schema['tables'].keys()
                if table not in tables_in_db
            ]
            if len(missing_tables) > 0:
                schemas_to_create.append((schema_network, schema_name))

    # print missing tables
    if len(schemas_to_create) == 0:
        print()
        print('All tables already exist')
        return
    else:
        print()
        print('Schemas to create:')
        for schema_network, schema_name in schemas_to_create:
            print('-', schema_name, '[network = ' + str(schema_network) + ']')

    # get confirmation
    if not confirm:
        print()
        if not toolcli.input_yes_or_no('continue? '):
            raise Exception('Skipping creation of tables')

    # create tables
    # (for now, use same database for all tables)

    # initialize schema versions schema if need be
    if not version_utils.is_schema_versions_initialized(conn=conn):
        initialize_schema_versions(conn=conn)

    # create each schema for each used network
    for schema_network, schema_name in schemas_to_create:
        initialize_schema(
            schema_name=schema_name,
            context=dict(network=schema_network),
            conn=conn,
        )

    print()
    print('All tables created')


def initialize_schema_versions(conn: toolsql.Connection) -> None:
    """initialize the schema_versions schema which manages versions of schemas"""

    initialize_schema('schema_versions', context=dict(network=None), conn=conn)


def initialize_schema(
    schema_name: spec.SchemaName,
    *,
    context: spec.Context,
    conn: toolsql.Connection,
) -> None:
    """initialize schema by creating its table and other objects"""

    network_schema_names = schema_utils.get_network_schema_names()
    prepared_schema = schema_name in network_schema_names

    # check that schema versions are being tracked
    if not version_utils.is_schema_versions_initialized(conn=conn):
        if schema_name != 'schema_versions':
            initialize_schema_versions(conn=conn)
    else:
        # check that schema not already initialized
        schema_version = version_utils.get_schema_version(
            schema_name=schema_name,
            context=context,
            conn=conn,
        )
        if schema_version is not None:
            raise Exception('schema already initialized')

    # load schema data
    if prepared_schema:
        schema = schema_utils.get_prepared_schema(
            schema_name=schema_name,
            context=context,
        )
    else:
        schema = schema_utils.get_raw_schema(schema_name=schema_name)

    # create tables
    for table_name, table_schema in schema['tables'].items():
        toolsql.create_table(
            table_schema,
            conn=conn,
            if_not_exists=True,
            confirm=True,
        )

    version_utils.set_schema_version(
        schema_name=schema_name,
        context=context,
        conn=conn,
    )


def drop_schema(
    schema_name: str,
    *,
    context: spec.Context = None,
    conn: toolsql.Connection,
    confirm: bool = False,
) -> None:

    if schema_name not in schema_utils.get_all_schema_names():
        raise Exception('unknown schema name: ' + str(schema_name))
    schema = typing.cast(db_types.SchemaName, schema_name)

    network: int | str | None = config.get_context_chain_id(context)
    if not confirm:
        if network is not None:
            answer = toolcli.input_yes_or_no(
                'Delete version row for schema '
                + schema
                + ' for network '
                + str(network)
                + '? '
            )
            if not answer:
                raise Exception('aborting')
        else:
            answer = toolcli.input_yes_or_no(
                'Delete ALL version rows for schema ' + schema + '? '
            )
            if not answer:
                raise Exception('aborting')

    if network is None:
        if schema in schema_utils.get_network_schema_names():
            networks: typing.Sequence[
                spec.NetworkReference | None
            ] = config.get_networks_that_have_providers()
        else:
            networks = [None]
    else:
        networks = [network]

    for network in networks:

        # determine tables in db
        db_table_names = toolsql.get_table_names(conn=conn)

        # determine tables in schema
        if network is not None:
            schema_data = schema_utils.get_prepared_schema(
                schema_name=schema,
                context=context,
            )
        else:
            schema_data = schema_utils.get_raw_schema(schema_name=schema)
        schema_table_names = schema_data['tables']

        # determine which tables to drop
        drop_table_names = set(db_table_names) & set(schema_table_names)

        # drop schema tables
        for table_name in drop_table_names:
            print('Dropping table', table_name)
            toolsql.drop_table(table=table_name, conn=conn, confirm=True)

    # delete rows from schema_versions table
    for network in networks:
        version_utils.delete_schema_version(
            schema_name=schema,
            network=network,
            conn=conn,
            confirm_delete_row=True,
            confirm_delete_schema=True,
        )




from __future__ import annotations

import toolcli

help_message = """run ctc setup wizard

setup wizard sets up ctc config and ctc data directory

each step is optional
- by default, setup process will leave existing settings unchanged
- setup can be rerun multiple times idempotently

can run in headless mode (non-interactive mode) using `--headless`:
- headless mode will use the default values for each unspecified parameter
- headless mode is useful in scripts when user input is not possible
- if config file already exists, headless mode must be run with `--overwrite`"""


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_setup_command,
        'help': help_message,
        'args': [
            {
                'name': '--headless',
                'help': 'run in headless mode',
                'action': 'store_true',
            },
            {
                'name': '--ignore-old-config',
                'help': 'ignore settings specified in previous config',
                'action': 'store_true',
            },
            {
                'name': '--skip-networking',
                'help': 'skip network configuration steps',
                'action': 'store_true',
            },
            {
                'name': '--rpc-url',
                'help': 'url to use for RPC node',
            },
            {
                'name': '--rpc-chain-id',
                'help': 'chain id of RPC node given by `--rpc-url`',
                'type': int,
            },
            {
                'name': '--data-dir',
                'help': 'directory where to store ctc data',
            },
            {
                'name': '--disable-logs',
                'help': 'disable ctc logs',
                'action': 'store_true',
            },
            {
                'name': '--skip-db',
                'help': 'skip creating db until it next time it is needed',
                'action': 'store_true',
            },
            {
                'name': '--overwrite',
                'help': 'overwrite old config file',
                'action': 'store_true',
            },
            {
                'name': '--skip-aliases',
                'help': 'disable ctc logs',
                'action': 'store_true',
            },
        ],
        'examples': {
            '': {'skip': True},
            '--headless': {'skip': True},
            '--headless --overwrite': {'skip': True},
        },
    }


async def async_setup_command(
    *,
    headless: bool,
    ignore_old_config: bool,
    skip_networking: bool,
    rpc_url: str | None,
    rpc_chain_id: int | None,
    data_dir: str,
    skip_db: bool,
    disable_logs: bool,
    overwrite: bool,
    skip_aliases: bool,
) -> None:
    from ctc.config import setup_utils

    await setup_utils.async_setup_ctc(
        headless=headless,
        ignore_old_config=ignore_old_config,
        skip_networking=skip_networking,
        rpc_url=rpc_url,
        rpc_chain_id=rpc_chain_id,
        data_dir=data_dir,
        disable_logs=disable_logs,
        skip_db=skip_db,
        overwrite=overwrite,
        skip_aliases=skip_aliases,
    )



from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec
from ... import rari_abis
from .. import pool_metadata
from .. import token_metadata


async def async_get_ctoken_exchange_rate(
    ctoken: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> int:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        function_abi=rari_abis.ctoken_function_abis['exchangeRateCurrent'],
        block_number=block,
        empty_token=None,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_ctoken_exchange_rate_by_block(
    ctoken: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
) -> list[int]:
    return await rpc.async_batch_eth_call(
        to_address=ctoken,
        function_abi=rari_abis.ctoken_function_abis['exchangeRateCurrent'],
        block_numbers=blocks,
        empty_token=None,
    )


async def async_get_ctoken_price(
    ctoken: spec.Address,
    *,
    oracle: spec.Address | None = None,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    raise_on_revert: bool = False,
) -> int | float:

    if oracle is None:
        oracle = await _async_get_ctoken_oracle(ctoken=ctoken)

    try:
        result = await rpc.async_eth_call(
            to_address=oracle,
            block_number=block,
            function_abi=rari_abis.oracle_function_abis['getUnderlyingPrice'],
            function_parameters=[ctoken],
        )
        if not isinstance(result, int):
            raise Exception('invalid rpc result')
        price: int | float = result
        if normalize:
            price /= 1e18
    except spec.RpcException as e:
        if raise_on_revert:
            raise e
        price = 0

    return price


async def _async_get_ctoken_oracle(ctoken: spec.Address) -> spec.Address:
    comptroller = await token_metadata.async_get_ctoken_comptroller(ctoken)
    oracle = await pool_metadata.async_get_pool_oracle(comptroller)
    return oracle



from __future__ import annotations

import toolcli


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': root_command,
        'help': 'cli interface to ctc and its subcommands',
        'args': [
            {
                'name': 'query',
                'nargs': '*',
                'help': 'address, block number, tx hash, or ERC20 symbol',
            }
        ],
        'examples': [
            '0x3a44deb6159843a584cf6ede188cd9f4161dc215',
            '14000000',
            '0xf48e1d4211c1df4cbf19aa0e99741ce0ad371a9906725bb60829147e763ffdf9',
            'DAI',
        ],
        'extra_data': ['parse_spec'],
    }


def root_command(query: str, parse_spec: toolcli.ParseSpec) -> None:

    execute_other_command_sequence = (
        toolcli.command_utils.execution.execute_other_command_sequence
    )

    if len(query) == 0:
        execute_other_command_sequence(
            command_sequence=('help',),
            args={'parse_spec': parse_spec},
            parse_spec=parse_spec,
        )
    elif len(query) == 1:
        item = query[0]
        if len(item) <= 9 and str.isnumeric(item):
            # treat as block number
            block = int(item)
            execute_other_command_sequence(
                command_sequence=('block',),
                args={'parse_spec': parse_spec, 'block': block},
                parse_spec=parse_spec,
            )

        elif item.startswith('0x') and len(item) == 66:
            # treat as transaction hash
            transaction = item
            execute_other_command_sequence(
                command_sequence=('tx',),
                args={'parse_spec': parse_spec, 'transaction': transaction},
                parse_spec=parse_spec,
            )

        elif (item.startswith('0x') and len(item) == 42) or item.endswith(
            '.eth'
        ):
            # treat as address
            address = item
            execute_other_command_sequence(
                command_sequence=('address',),
                args={'parse_spec': parse_spec, 'address': address},
                parse_spec=parse_spec,
            )

        elif str.isalnum(item) and len(item) <= 20:
            try:
                execute_other_command_sequence(
                    command_sequence=('symbol',),
                    args={'parse_spec': parse_spec, 'query': item},
                    parse_spec=parse_spec,
                )
            except LookupError:
                execute_other_command_sequence(
                    command_sequence=('help',),
                    args={'parse_spec': parse_spec},
                    parse_spec=parse_spec,
                )

        else:
            raise Exception('unknown query: ' + str(query))
    else:
        raise Exception('unknown query: ' + str(query))



from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc.protocols import uniswap_v2_utils
from ctc.defi.dex_utils.amm_utils import cpmm
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_pool_command,
        'help': 'summarize pool',
        'args': [
            {'name': 'pool', 'help': 'pool address'},
            {
                'name': '--block',
                'default': 'latest',
                'help': 'block number range',
            },
            {'name': '--depths', 'help': 'liquidity depths', 'nargs': '+'},
        ],
        'examples': [
            '0xae461ca67b15dc8dc81ce7615e0320da1a9ab8d5 --block 14000000',
        ],
    }


async def async_pool_command(
    *,
    pool: spec.Address,
    block: str,
    depths: typing.Sequence[str] | None,
) -> None:
    tokens_metadata = await uniswap_v2_utils.async_get_pool_tokens_metadata(
        pool
    )
    x_symbol = tokens_metadata['x_symbol']
    y_symbol = tokens_metadata['y_symbol']
    title = 'Uniswap V2 Pool: ' + x_symbol + ' x ' + y_symbol
    toolstr.print_text_box(title)
    print('-', x_symbol, 'address:', tokens_metadata['x_address'])
    print('-', y_symbol, 'address:', tokens_metadata['y_address'])
    print('-', x_symbol, 'decimals:', tokens_metadata['x_decimals'])
    print('-', y_symbol, 'decimals:', tokens_metadata['y_decimals'])
    print()
    print()

    if depths is not None:
        depths_float: list[float] | None = [float(depth) for depth in depths]
    else:
        depths_float = None

    toolstr.print_header('Pool State')
    pool_state = await uniswap_v2_utils.async_get_pool_state(pool, block=block)
    cpmm.print_pool_summary(
        x_name=x_symbol,
        y_name=y_symbol,
        depths=depths_float,
        **pool_state,
    )



from __future__ import annotations

import os


config_path_env_var = 'CTC_CONFIG_PATH'
provider_env_var = 'CTC_PROVIDER'
network_env_var = 'CTC_NETWORK'
cache_env_var = 'CTC_CACHE'

allowed_config_filetypes = ['.json']

default_config_path = os.path.expanduser('~/.config/ctc/config.json')

min_allowed_config_version = (0, 2, 10)
min_recommended_config_version = (0, 3, 0)



"""functions for getting dex classes"""

from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from . import dex_directory
from . import dex_class


def get_all_dex_classes() -> typing.Mapping[str, typing.Type[dex_class.DEX]]:
    """get all DEX classes"""
    dex_names = [
        'balancer',
        'curve',
        'sushi',
        'uniswapv2',
        'uniswapv3',
    ]
    return {dex_name: get_dex_class(dex_name) for dex_name in dex_names}


def get_dex_class(
    dex: typing.Type[dex_class.DEX] | str | None = None,
    *,
    factory: spec.Address | None = None,
    context: spec.Context = None,
) -> typing.Type[dex_class.DEX]:
    """return DEX object corresponding to dex name or dex factory"""

    if dex is not None:
        if isinstance(dex, str):
            return _get_dex_class_from_name(dex)
        elif issubclass(dex, dex_class.DEX):
            return dex
        else:
            raise Exception('unknown dex type: ' + str(type(dex)))
    elif factory is not None:
        return _get_dex_class_from_factory(factory=factory, context=context)
    else:
        raise Exception('not enough inputs specified')


async def async_get_dex_class(
    dex: typing.Type[dex_class.DEX] | str | None = None,
    *,
    factory: spec.Address | None = None,
    pool: spec.Address | None = None,
    context: spec.Context = None,
) -> typing.Type[dex_class.DEX]:
    """get DEX class matching given inputs"""

    if factory is not None or dex is not None:
        return get_dex_class(
            dex=dex,
            factory=factory,
            context=context,
        )
    else:
        if pool is None:
            raise Exception(
                'must specify dex, factory or pool to get dex class'
            )
        return await _async_get_dex_class_of_pool(pool, context=context)


def _get_dex_class_from_factory(
    factory: spec.Address,
    context: spec.Context = None,
) -> typing.Type[dex_class.DEX]:
    """return DEX class using given DEX factory"""
    dex_name = dex_directory.get_dex_name_of_factory(factory, context=context)
    return _get_dex_class_from_name(dex_name)


def _get_dex_class_from_name(dex: str) -> typing.Type[dex_class.DEX]:
    """return DEX class using given DEX name"""

    dex = dex.lower().replace(' ', '').replace('-', '').replace('_', '')

    if dex == 'balancer':
        from .dex_implementations import balancer_dex

        return balancer_dex.BalancerDEX
    elif dex == 'curve':
        from .dex_implementations import curve_dex

        return curve_dex.CurveDEX
    elif dex == 'sushi':
        from .dex_implementations import sushi_dex

        return sushi_dex.SushiDEX
    elif dex == 'uniswapv2':
        from .dex_implementations import uniswap_v2_dex

        return uniswap_v2_dex.UniswapV2DEX
    elif dex == 'uniswapv3':
        from .dex_implementations import uniswap_v3_dex

        return uniswap_v3_dex.UniswapV3DEX
    else:
        raise Exception('unknown DEX: ' + str(dex))


async def _async_get_dex_class_of_pool(
    pool: spec.Address,
    context: spec.Context = None,
) -> typing.Type[dex_class.DEX]:
    """get DEX class for given pool"""

    from ctc import db

    dex_pool = await db.async_query_dex_pool(address=pool, context=context)
    if dex_pool is None:
        raise Exception('could not determine dex class of pool')
    return _get_dex_class_from_factory(dex_pool['factory'], context=context)




from __future__ import annotations

import asyncio
import typing
import shutil

import ctc.rpc
from ctc import spec
import tooljob

try:
    import pdp
except ImportError:
    raise Exception('install pdp for dataset functionality')


class ExtractErc20Transfers(pdp.BlockChunkJobs):
    tracker: tooljob.trackers.file_tracker.FileTracker

    def execute_job(self, i: int) -> typing.Any:
        job_data = self.get_job_data(i)
        job_name = self.get_job_name(i)
        path = self.tracker.get_job_output_path(i)
        sync_extract_erc20_transfers(
            start_block=job_data['start_block'],
            end_block=job_data['end_block'],
            job_name=job_name,
            path=path,
            context=self.context,
        )


def sync_extract_erc20_transfers(
    *,
    start_block: int,
    end_block: int,
    job_name: str,
    path: str,
    context: spec.Context,
) -> None:
    try:
        asyncio.run(
            async_extract_erc20_transfers(
                start_block=start_block,
                end_block=end_block,
                path=path,
                context=context,
            )
        )
    except Exception as e:
        print('job', job_name, 'failed:' + str(e))
        raise e


async def async_extract_erc20_transfers(
    *,
    start_block: int,
    end_block: int,
    path: str,
    context: spec.Context,
) -> None:

    event_hash = '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'
    encoded_transfers = await ctc.async_get_events(
        event_hash=event_hash,
        start_block=start_block,
        end_block=end_block,
        decode=False,
        context={'cache': False},
        verbose=False,
        max_blocks_per_request=100,
    )

    await ctc.rpc.async_close_http_session(context=context)

    # load data into output file
    decoded_transfers = (
        ctc.evm.erc20_utils.erc20_events._decode_erc20_transfers(
            encoded_transfers
        )
    )
    del encoded_transfers

    # write file
    temp_path = path + '_temp'
    if len(decoded_transfers) == 0:
        decoded_transfers.write_parquet(temp_path)
    else:
        decoded_transfers.write_parquet(
            temp_path,
            statistics=True,
            row_group_size=max(1000, int(len(decoded_transfers) / 64)),
        )
    shutil.move(temp_path, path)




from __future__ import annotations

import toolsql


contract_abis_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'contract_abis': {
            'columns': [
                {'name': 'address', 'type': 'Text', 'primary': True},
                {'name': 'abi_text', 'type': 'Text'},
                {'name': 'includes_proxy', 'type': 'Text'},
            ],
        },
    },
}



"""

## Transaction Types
0. legacy
1. [EIP-1559](https://eips.ethereum.org/EIPS/eip-1559)
2. [EIP-2930](https://eips.ethereum.org/EIPS/eip-2930)
"""
from __future__ import annotations

import typing

from . import address_types
from . import log_types
from . import binary_types

from typing_extensions import TypedDict, NotRequired


TransactionHash = binary_types.PrefixHexData

TransactionAccessList = typing.Sequence[
    # tuples describe accessed data as (contract, memory_slot)
    typing.Tuple[
        address_types.Address,
        typing.Sequence[binary_types.PrefixHexData],
    ]
]


#
# # pre-chain transaction types
#

# these transaction types are either sign-able or submit-able

class LegacyTransaction(TypedDict):
    #
    # used for pre-sign hash
    nonce: int
    gas_price: int
    gas: int
    to: address_types.Address
    value: int
    input: binary_types.Data
    #
    # included once sigend
    v: NotRequired[binary_types.Data]
    r: NotRequired[binary_types.Data]
    s: NotRequired[binary_types.Data]
    #
    # other data
    chain_id: NotRequired[binary_types.Data]
    type: int


class EIP2930Transaction(TypedDict):
    #
    # used for pre-sign hash
    chain_id: int
    nonce: int
    gas_price: int
    gas: int
    to: address_types.Address
    value: int
    input: binary_types.Data
    access_list: TransactionAccessList
    #
    # included once sigend
    v: NotRequired[binary_types.Data]
    r: NotRequired[binary_types.Data]
    s: NotRequired[binary_types.Data]
    #
    # other data
    type: NotRequired[binary_types.Data]


class EIP1559Transaction(TypedDict):
    #
    # used for pre-sign hash
    chain_id: int
    nonce: int
    max_priority_fee_per_gas: int
    max_fee_per_gas: int
    gas: int
    to: address_types.Address
    value: int
    input: binary_types.Data
    access_list: TransactionAccessList
    #
    # included once sigend
    v: NotRequired[binary_types.Data]
    r: NotRequired[binary_types.Data]
    s: NotRequired[binary_types.Data]
    #
    # other data
    type: NotRequired[binary_types.Data]


PrechainTransaction = typing.Union[
    LegacyTransaction,
    EIP2930Transaction,
    EIP1559Transaction,
]


#
# # RPC transaction types
#

# these transaction types are the results of RPC requests

RPCTransaction = TypedDict(
    'RPCTransaction',
    {
        'access_list': NotRequired[TransactionAccessList],
        'block_hash': binary_types.PrefixHexData,
        'block_number': int,
        'chain_id': binary_types.PrefixHexData,
        'from': address_types.Address,
        'gas': int,
        'gas_price': int,
        'hash': TransactionHash,
        'input': binary_types.PrefixHexData,
        'max_fee_per_gas': NotRequired[int],
        'max_priority_fee_per_gas': NotRequired[int],
        'nonce': int,
        'r': binary_types.PrefixHexData,
        's': binary_types.PrefixHexData,
        'to': address_types.Address,
        'transaction_index': int,
        'type': int,
        'v': int,
        'value': int,
    },
)

RPCTransactionReceipt = TypedDict(
    'RPCTransactionReceipt',
    {
        'block_hash': str,
        'block_number': int,
        'contract_address': str,
        'cumulative_gas_used': int,
        'effective_gas_price': int,
        'from': str,
        'gas_used': int,
        'logs': typing.Sequence[log_types.RawLog],
        'logs_bloom': str,
        'status': int,
        'to': str,
        'transaction_hash': str,
        'transaction_index': int,
        'type': int,
    },
)


#
# # DB transaction types
#

# these transaction types are stored in the ctc DB

class DBTransaction(TypedDict):
    hash: TransactionHash
    block_number: int
    transaction_index: int
    to_address: address_types.Address
    from_address: address_types.Address
    value: int
    input: str
    nonce: int
    transaction_type: int
    status: int
    gas_used: int
    gas_limit: int
    gas_priority: int | None
    gas_price: int
    gas_price_max: int | None
    access_list: TransactionAccessList | None


# convert large fields to str
class DBTransactionText(TypedDict):
    hash: TransactionHash
    block_number: int
    transaction_index: int
    to_address: address_types.Address
    from_address: address_types.Address
    value: str
    input: str
    nonce: int
    transaction_type: int
    status: int
    gas_used: int
    gas_limit: int
    gas_priority: int | None
    gas_price: int
    gas_price_max: int | None
    access_list: TransactionAccessList | None




"""utilitize for config file IO"""

from __future__ import annotations

import os
import functools
import sys
import typing

import ctc
from ctc import spec
from . import config_env_vars
from . import config_overrides
from . import config_spec
from . import config_validate
from . import upgrade_utils


def get_config_path(*, raise_if_dne: bool = True) -> str:

    config_path_env_var = config_spec.config_path_env_var
    default_config_path = config_spec.default_config_path

    # get config_path from environmental variable
    if config_path_env_var is not None:
        config_path = os.environ.get(config_path_env_var)
        if config_path == '':
            config_path = None

    # use default config path if not specified
    if config_path is None and default_config_path is not None:
        config_path = default_config_path

    # validate config_path
    if config_path is None:
        raise spec.ConfigPathNotSet('config path is not set')
    else:
        if raise_if_dne and not os.path.isfile(config_path):
            raise spec.ConfigDoesNotExist(
                'config at path does not exist: ' + str(config_path)
            )

    return config_path


@functools.lru_cache()
def get_config(
    warn_if_dne: bool = True,
    warn_if_outdated: bool = True,
) -> spec.Config:

    # load from file
    try:
        config_path = get_config_path()
        with open(config_path, 'r') as f:
            import json

            raw_config = json.load(f)
    except spec.ConfigDoesNotExist:
        from . import config_defaults

        if warn_if_dne:
            print(
                '[WARNING]'
                ' ctc config file does not exist;'
                ' use `ctc setup` on command line to generate a config file',
                file=sys.stderr,
            )
        raw_config = config_defaults.get_default_config(use_env_variables=True)

    # auto-upgrade config if need be
    config_version = raw_config.get('config_spec_version')
    if config_version is not None:
        config_stable_version = upgrade_utils.get_stable_version(config_version)
    else:
        config_stable_version = None
    ctc_stable_version = upgrade_utils.get_stable_version(ctc.__version__)
    if config_stable_version != ctc_stable_version:
        if warn_if_outdated:
            print(
                '[WARNING] using outdated config -- run `ctc setup` on command line to update',
                file=sys.stderr,
            )
        raw_config = upgrade_utils.upgrade_config(raw_config)

    # convert int keys from str to int
    for key in spec.typedata.config_int_subkeys:
        if raw_config.get(key) is not None:
            raw_config[key] = {
                int(chain_id): network_metadata
                for chain_id, network_metadata in raw_config[key].items()
            }

    # load settings from env vars
    raw_config = config_env_vars._add_config_env_vars(raw_config)

    # add config overrides
    raw_config = raw_config
    overrides = config_overrides.get_config_overrides()
    if len(overrides) > 0:
        raw_config = dict(raw_config)
        raw_config.update(overrides)

    # validate
    config_validate.validate_config(raw_config)
    return raw_config  # type: ignore


def get_config_version_tuple(
    config: typing.Mapping[str, typing.Any],
) -> typing.Tuple[int, int, int]:

    if 'config_spec_version' in config:
        version_str = config['config_spec_version']
        if isinstance(version_str, str):
            version_tuple = tuple(
                int(token) for token in version_str.split('.')
            )
            if len(version_tuple) == 3:
                return (version_tuple[0], version_tuple[1], version_tuple[2])

    elif 'config_version' in config:
        config_version = config['config_version']
        if (
            isinstance(config_version, list)
            and len(config_version) == 3
            and isinstance(config_version[0], int)
            and isinstance(config_version[1], int)
            and isinstance(config_version[2], int)
        ):
            return (config_version[0], config_version[1], config_version[2])

    raise Exception('could not detect config version')


def reset_config_cache() -> None:
    get_config.cache_clear()




from __future__ import annotations

import typing


if typing.TYPE_CHECKING:
    import numpy
    import polars as pl

    DataFrame = pl.DataFrame
    Series = pl.Series
    DType = numpy.typing.DTypeLike
    NumpyArray = numpy.typing.NDArray  # type: ignore

    IntegerOutputFormatScalar = typing.Union[
        type[pl.datatypes.DataTypeClass],
        type[object],
        type[int],
        type[float],
        typing.Literal['decimal'],
    ]
    IntegerOutputFormat = typing.Union[
        IntegerOutputFormatScalar,
        typing.Mapping[str, IntegerOutputFormatScalar],
    ]

else:
    DataFrame = typing.Any
    Series = typing.Any
    DType = typing.Any
    NumpyArray = typing.Any


Integer = typing.Union[
    int,
    'numpy.int8',
    'numpy.int16',
    'numpy.int32',
    'numpy.int64',
]

Float = typing.Union[
    float,
    'numpy.float16',
    'numpy.float32',
    'numpy.float64',
    'numpy.float128',
]

Number = typing.Union[Integer, Float]




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import tooltime

from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.protocols import uniswap_v2_utils
from .. import dex_class


class UniswapV2DEX(dex_class.DEX):
    """Uniswap V2 DEX"""

    _pool_factories = {1: [uniswap_v2_utils.uniswap_v2_factory]}

    @classmethod
    async def async_get_new_pools(
        cls,
        *,
        factory: spec.Address,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:
        df = await evm.async_get_events(
            factory,
            event_abi=uniswap_v2_utils.factory_event_abis['PairCreated'],
            verbose=False,
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            context=context,
        )

        dex_pools = []
        for row in df.to_dicts():
            dex_pool: spec.DexPool = {
                'address': row['arg__pair'],
                'factory': factory,
                'asset0': row['arg__token0'],
                'asset1': row['arg__token1'],
                'asset2': None,
                'asset3': None,
                'fee': int(0.003 * 1e8),
                'creation_block': row['block_number'],
                'additional_data': {},
            }
            dex_pools.append(dex_pool)
        return dex_pools

    @classmethod
    async def _async_get_pool_assets_from_node(
        cls,
        pool: spec.Address,
        *,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> tuple[str, str]:
        import asyncio
        from ctc.protocols import uniswap_v2_utils

        token0 = rpc.async_eth_call(
            function_abi=uniswap_v2_utils.pool_function_abis['token0'],
            to_address=pool,
            block_number=block,
            context=context,
        )
        token1 = rpc.async_eth_call(
            function_abi=uniswap_v2_utils.pool_function_abis['token1'],
            to_address=pool,
            block_number=block,
            context=context,
        )

        return await asyncio.gather(token0, token1)

    @classmethod
    async def _async_get_pool_raw_trades(
        cls,
        pool: spec.Address,
        *,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        include_timestamps: bool = False,
        context: spec.Context = None,
        verbose: bool = False,
    ) -> spec.RawDexTrades:
        import polars as pl

        trades = await evm.async_get_events(
            event_abi=uniswap_v2_utils.pool_event_abis['Swap'],
            contract_address=pool,
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            include_timestamps=include_timestamps,
            verbose=verbose,
            context=context,
            integer_output_format=float,
        )

        df = trades.select(
            (pl.when(pl.col('arg__amount0Out') > 0).then(1).otherwise(0)).alias(
                'sold_id'
            ),
            (pl.col('arg__amount0In') + pl.col('arg__amount1In')).alias(
                'sold_amount'
            ),
            (pl.col('arg__amount0Out') + pl.col('arg__amount1Out')).alias(
                'bought_amount'
            ),
        )
        df = df.with_columns((1 - pl.col('sold_id')).alias('bought_id'))

        output: spec.RawDexTrades = {
            'block_number': trades['block_number'],
            'transaction_hash': trades['transaction_hash'],
            'recipient': trades['arg__to'],
            'sold_id': df['sold_id'],
            'bought_id': df['bought_id'],
            'sold_amount': df['sold_amount'],
            'bought_amount': df['bought_amount'],
        }

        if include_timestamps:
            output['timestamp'] = trades['timestamp']

        return output




"""
see https://docs.aave.com/developers/v/2.0/guides/apy-and-apr

incentives not currently included
"""
from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.toolbox import nested_utils
from . import aave_pool_tokens

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class AaveV2ReserveData(TypedDict):
        configuration: int
        liquidity_index: int
        variable_borrow_index: int
        current_liquidity_rate: int
        current_variable_borrow_rate: int
        current_stable_borrow_rate: int
        last_update_timestamp: int
        atoken_address: spec.Address
        stable_debt_token_address: spec.Address
        variable_debt_token_address: spec.Address
        interest_rate_strategy_address: spec.Address
        id: int

    class AaveV2ReserveListData(TypedDict):
        configuration: list[int]
        liquidity_index: list[int]
        variable_borrow_index: list[int]
        current_liquidity_rate: list[int]
        current_variable_borrow_rate: list[int]
        current_stable_borrow_rate: list[int]
        last_update_timestamp: list[int]
        atoken_address: list[spec.Address]
        stable_debt_token_address: list[spec.Address]
        variable_debt_token_address: list[spec.Address]
        interest_rate_strategy_address: list[spec.Address]
        id: list[int]

    class AaveV2TokenMarket(TypedDict):
        underlying: spec.Address
        symbol: str
        reserve_data: AaveV2ReserveData


ray = 10**27
seconds_per_year = 31536000

aave_lending_pool = '0x7d2768de32b0b80b7a3454c06bdac94a69ddc7a9'


async def async_get_reserve_data(
    asset: spec.Address,
    block: spec.BlockNumberReference | None = None,
    *,
    context: spec.Context = None,
) -> AaveV2ReserveData:
    result = await rpc.async_eth_call(
        to_address=aave_lending_pool,
        function_name='getReserveData',
        function_parameters=[asset],
        block_number=block,
        context=context,
    )

    return {
        'configuration': result[0],
        'liquidity_index': result[1],
        'variable_borrow_index': result[2],
        'current_liquidity_rate': result[3],
        'current_variable_borrow_rate': result[4],
        'current_stable_borrow_rate': result[5],
        'last_update_timestamp': result[6],
        'atoken_address': result[7],
        'stable_debt_token_address': result[8],
        'variable_debt_token_address': result[9],
        'interest_rate_strategy_address': result[10],
        'id': result[11],
    }


async def async_get_reserve_data_by_block(
    asset: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    context: spec.Context = None,
) -> AaveV2ReserveListData:
    coroutines = [
        async_get_reserve_data(asset, block=block, context=context)
        for block in blocks
    ]

    results = await asyncio.gather(*coroutines)
    return nested_utils.list_of_dicts_to_dict_of_lists(results)  # type: ignore


async def async_get_reserves_data(
    *,
    reserves_list: typing.Sequence[spec.Address] | None = None,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> typing.Sequence[AaveV2ReserveData]:
    if reserves_list is None:
        reserves_list = await aave_pool_tokens.async_get_reserves_list(
            block=block,
            context=context,
        )

    results = await rpc.async_batch_eth_call(
        to_address=aave_lending_pool,
        function_name='getReserveData',
        function_parameter_list=[[asset] for asset in reserves_list],
        block_number=block,
        context=context,
    )

    return [
        {
            'configuration': result[0],
            'liquidity_index': result[1],
            'variable_borrow_index': result[2],
            'current_liquidity_rate': result[3],
            'current_variable_borrow_rate': result[4],
            'current_stable_borrow_rate': result[5],
            'last_update_timestamp': result[6],
            'atoken_address': result[7],
            'stable_debt_token_address': result[8],
            'variable_debt_token_address': result[9],
            'interest_rate_strategy_address': result[10],
            'id': result[11],
        }
        for result in results
    ]


async def async_get_token_markets(
    *,
    reserves_list: typing.Sequence[spec.Address] | None = None,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> typing.Sequence[AaveV2TokenMarket]:
    if reserves_list is None:
        reserves_list = await aave_pool_tokens.async_get_reserves_list(
            block=block, context=context
        )

    reserves_data_task = asyncio.create_task(
        async_get_reserves_data(
            reserves_list=reserves_list, block=block, context=context
        )
    )
    symbols_task = asyncio.create_task(
        evm.async_get_erc20s_symbols(reserves_list, context=context)
    )

    reserves_data = await reserves_data_task
    symbols = await symbols_task

    markets = []
    for r in range(len(reserves_list)):
        market: AaveV2TokenMarket = {
            'underlying': reserves_list[r],
            'symbol': symbols[r],
            'reserve_data': reserves_data[r],
        }
        markets.append(market)
    return markets


async def async_get_interest_rates(
    *,
    token: spec.Address | None = None,
    block: spec.BlockNumberReference | None = None,
    reserve_data: AaveV2ReserveData | None = None,
    context: spec.Context = None,
) -> dict[str, float]:
    if reserve_data is None:
        if token is None:
            raise Exception('must specify token or reserve_data')
        reserve_data = await async_get_reserve_data(
            asset=token, block=block, context=context
        )

    supply_apr = reserve_data['current_liquidity_rate'] / ray
    supply_apy = (1 + supply_apr / seconds_per_year) ** seconds_per_year - 1
    borrow_apr = reserve_data['current_variable_borrow_rate'] / ray
    borrow_apy = (1 + borrow_apr / seconds_per_year) ** seconds_per_year - 1

    return {
        'supply_apr': supply_apr,
        'supply_apy': supply_apy,
        'borrow_apr': borrow_apr,
        'borrow_apy': borrow_apy,
    }


async def async_get_interest_rates_by_block(
    token: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    *,
    reserve_data_by_block: AaveV2ReserveListData | None = None,
    context: spec.Context = None,
) -> dict[str, list[float]]:
    import numpy as np

    if reserve_data_by_block is None:
        reserve_data_by_block = await async_get_reserve_data_by_block(
            asset=token, blocks=blocks, context=context
        )

    currrent_liquidity_rate: spec.NumpyArray = np.array(
        reserve_data_by_block['current_liquidity_rate']
    )
    current_variable_borrow_rate: spec.NumpyArray = np.array(
        reserve_data_by_block['current_variable_borrow_rate']
    )

    supply_apr = currrent_liquidity_rate / ray
    supply_apy = (1 + supply_apr / seconds_per_year) ** seconds_per_year - 1
    borrow_apr = current_variable_borrow_rate / ray
    borrow_apy = (1 + borrow_apr / seconds_per_year) ** seconds_per_year - 1

    return {
        'supply_apr': list(supply_apr),
        'supply_apy': list(supply_apy),
        'borrow_apr': list(borrow_apr),
        'borrow_apy': list(borrow_apy),
    }




from __future__ import annotations

import typing

import toolcli
import toolstr
import toolsql

from ctc import cli
from ctc import config
from ctc import db
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': status_command,
        'help': 'display database status',
        'args': [
            {
                'name': ['-v', '--verbose'],
                'help': 'display additional information',
                'action': 'store_true',
            },
        ],
        'examples': [''],
    }


def status_command(verbose: bool) -> None:
    styles = cli.get_cli_styles()

    toolstr.print_text_box('Database Status', style=styles['title'])
    print()
    context: spec.Context = {}
    db_config = config.get_context_db_config(
        context=context, schema_name='schema_versions'
    )
    if db_config is None:
        print('[no db configured]')
        return

    toolstr.print_header('Database Config', style=styles['title'])
    for key, value in db_config.items():
        if key == 'password':
            value = '********'
        cli.print_bullet(key=key, value=value)

    # print data being stored
    active_schemas = db.get_active_schemas()
    print()
    toolstr.print_header('Data to Store', style=styles['title'])
    networks = config.get_networks_that_have_providers()
    admin_schemas = db.get_admin_schema_names()
    generic_schemas = db.get_generic_schema_names()
    network_schemas = db.get_network_schema_names()
    toolstr.print('- admin schemas', style=styles['option'])
    for admin_schema in admin_schemas:
        toolstr.print('    -', admin_schema, style=styles['description'])
    toolstr.print('- generic schemas', style=styles['option'])
    for generic_schema in generic_schemas:
        toolstr.print('    -', generic_schema, style=styles['description'])
    toolstr.print('- evm schemas', style=styles['option'])
    for datatype in network_schemas:
        if active_schemas.get(datatype):
            toolstr.print('    -', datatype, style=styles['description'])
        else:
            toolstr.print(
                '    -', datatype, '(inactive)', style=styles['description']
            )
    toolstr.print('- networks', style=styles['option'])
    for network in networks:
        toolstr.print('    -', network, style=styles['description'])

    print()
    db_exists = toolsql.does_db_exist(db_config)
    if db_exists:

        with toolsql.connect(db_config) as conn:
            db_schema = toolsql.get_db_schema(conn)

        with toolsql.connect(db_config) as conn:
            if verbose:
                # toolsql.print_db_schema(db_schema=db_schema, styles=styles)

                print()
                toolstr.print_header('Schema Versions', style=styles['title'])
                rows = []
                for schema_name in active_schemas:
                    if schema_name in network_schemas:
                        schema_networks: typing.Sequence[
                            spec.NetworkReference
                        ] | typing.Sequence[
                            None
                        ] = config.get_networks_that_have_providers()
                    else:
                        schema_networks = [None]
                    for schema_network in schema_networks:
                        version = db.get_schema_version(
                            schema_name,
                            context=dict(network=schema_network),
                            conn=conn,
                        )
                        if version is None:
                            version = '[DNE]'

                        if schema_network is None:
                            network_str = ''
                        else:
                            network_str = str(schema_network)
                        row = [schema_name, network_str, version]
                        rows.append(row)
                    rows = sorted(rows, key=lambda row: tuple(row))
                toolstr.print_table(
                    rows,
                    labels=['schema', 'network', 'version'],
                    indent=4,
                    border=styles['comment'],
                    label_style=styles['title'],
                    column_styles={
                        'schema': styles['description'],
                        'network': 'bold',
                        'version': 'bold',
                    },
                )

                print()
                print()
                toolstr.print_header('Table Usage', style=styles['title'])
                toolsql.print_db_usage(conn, styles=styles)

    else:
        toolstr.print_text_box('Schema Summary', style=styles['title'])
        print('- db does not exist')




from __future__ import annotations

import typing

from ctc import spec
from ... import binary_utils
from .. import contract_abi_utils
from . import function_abi_parsing


def get_function_abi(
    contract_abi: spec.ContractABI,
    function_name: typing.Optional[str] = None,
    *,
    n_parameters: typing.Optional[int] = None,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
    function_selector: typing.Optional[spec.FunctionSelector] = None,
) -> spec.FunctionABI:
    """get function ABI from contract ABI"""

    if function_selector is not None:
        function_selector = binary_utils.to_hex(function_selector)

    candidates = []
    for item in contract_abi:
        if item.get('type') != 'function':
            continue
        else:

            if typing.TYPE_CHECKING:
                function_abi = typing.cast(spec.FunctionABI, item)
            else:
                function_abi = item

        if (
            function_name is not None
            and function_abi.get('name') != function_name
        ):
            continue
        if (
            n_parameters is not None
            and len(function_abi['inputs']) != n_parameters
        ):
            continue
        if parameter_types is not None:
            types = function_abi_parsing.get_function_parameter_types(
                function_abi
            )
            if tuple(parameter_types) != tuple(types):
                continue
        if function_selector is not None:
            item_selector = function_abi_parsing.get_function_selector(
                function_abi
            )
            item_selector = binary_utils.to_hex(item_selector)
            if item_selector != function_selector:
                continue
        candidates.append(function_abi)

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) == 0:
        raise LookupError('could not find function abi')
    elif len(candidates) > 0:
        import json

        as_json = [json.dumps(candidate) for candidate in candidates]
        if all(as_json[0] == entry for entry in as_json[1:]):
            return candidates[0]

        raise LookupError('too many candidates found for function abi')
    else:
        raise Exception('internal error')


def sync_get_function_abi(
    *,
    function_name: typing.Optional[str] = None,
    contract_abi: typing.Optional[spec.ContractABI] = None,
    contract_address: typing.Optional[spec.Address] = None,
    n_parameters: typing.Optional[int] = None,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
    function_selector: typing.Optional[spec.FunctionSelector] = None,
    context: spec.Context = None,
) -> spec.FunctionABI:
    """get function abi from ABI source"""
    raise NotImplementedError('sync version of get_function_abi')
    # necessary functions to replace:
    # - contract_utils.async_get_proxy_implementation
    # - etherscan_utils.async_get_contract_abi
    # - db.async_query_contract_ab
    # - db.async_intake_contract_abi


async def async_get_function_abi(
    *,
    function_name: typing.Optional[str] = None,
    contract_abi: typing.Optional[spec.ContractABI] = None,
    contract_address: typing.Optional[spec.Address] = None,
    n_parameters: typing.Optional[int] = None,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
    function_selector: typing.Optional[spec.FunctionSelector] = None,
    context: spec.Context = None,
) -> spec.FunctionABI:
    """get function ABI from local database or block explorer"""

    if contract_abi is None:
        if contract_address is None:
            raise Exception('must specify contract_abi or contract_address')
        contract_abi = await contract_abi_utils.async_get_contract_abi(
            contract_address=contract_address,
            context=context,
        )

    try:
        return get_function_abi(
            function_name=function_name,
            contract_abi=contract_abi,
            n_parameters=n_parameters,
            parameter_types=parameter_types,
            function_selector=function_selector,
        )

    except LookupError as e:

        # query contract_abi again if contract abi might have changed since db
        if contract_address is not None:
            from ctc import config

            context = config.update_context(
                context=context, cache={'read': False}
            )
            contract_abi = await contract_abi_utils.async_get_contract_abi(
                contract_address=contract_address,
                context=context,
            )

            return get_function_abi(
                function_name=function_name,
                contract_abi=contract_abi,
                n_parameters=n_parameters,
                parameter_types=parameter_types,
                function_selector=function_selector,
            )

        else:
            raise e


def get_function_abis(
    contract_abi: spec.ContractABI,
) -> typing.Sequence[spec.FunctionABI]:
    """get list of function ABI's in contract ABI"""
    return [item for item in contract_abi if item['type'] == 'function']




"""
## Resources
for overview of different chains see
- https://github.com/ethereum-lists/chains
- https://chainlist.org/
"""

from __future__ import annotations

import typing

from ctc import spec


def get_network_name(network: spec.NetworkName | int) -> spec.NetworkName:
    """get name of network"""

    if isinstance(network, str):
        return network

    config_network_names_by_id = _get_network_names_by_chain_id()
    if network in config_network_names_by_id:
        name = config_network_names_by_id[network]
        if name is None:
            raise LookupError('network name is not known')
        elif isinstance(name, str):
            return name
        else:
            raise Exception('bad type for name: ' + str(type(name)))
    else:
        raise LookupError('unknown network: ' + str(network))


def get_network_chain_id(network: spec.NetworkName | int) -> spec.ChainId:
    """get chain id of network"""

    if isinstance(network, int):
        return network

    config_chain_ids_by_network_name = _get_chain_ids_by_network_name()
    if network in config_chain_ids_by_network_name:
        return config_chain_ids_by_network_name[network]
    else:
        raise LookupError('unknown network: ' + str(network))


def _get_chain_ids_by_network_name() -> typing.Mapping[spec.NetworkName, int]:
    return {
        network_metadata['name']: chain_id
        for chain_id, network_metadata in get_networks().items()
        if network_metadata['name'] is not None
    }


def get_networks() -> typing.Mapping[int, spec.NetworkMetadata]:
    """get networks of current configuation"""

    from ctc import config

    return config.get_config_networks()


def _get_network_names_by_chain_id() -> typing.Mapping[
    int, spec.NetworkName | None
]:
    return {
        chain_id: network_metadata['name']
        for chain_id, network_metadata in get_networks().items()
    }


def get_network_metadata(
    network: spec.NetworkReference,
) -> spec.NetworkMetadata:
    """get metadata of network"""

    if isinstance(network, str):
        network = get_network_chain_id(network)

    networks = get_networks()
    if network in networks:
        return networks[network]
    else:
        raise LookupError('unknown network: ' + str(network))


def get_network_block_explorer(network: spec.NetworkReference) -> str | None:
    """get block explorer for network"""

    network_metadata = get_network_metadata(network)
    return network_metadata['block_explorer']




from __future__ import annotations

import typing
from ctc import evm
from ctc import spec


#
# # addresses
#

tick_lens = '0xbfd8137f7d1516d3ea5ca83523914859ec47f573'
quoter = '0xb27308f9f90d607463bb33ea1bebb41c27ce5ab6'
factory = '0x1f98431c8ad98523631ae4a59f267346ea31f984'


#
# # abi's
#


_abi_cache: dict[
    str,
    typing.Optional[
        dict[
            str,
            typing.Union[
                spec.FunctionABI,
                spec.EventABI,
                spec.ErrorABI,
            ],
        ]
    ],
] = {
    'pool': None,
    'quoter': None,
    'tick_lens': None,
}


async def _async_load_abi_cache(contract: str) -> None:

    # ensure cache in empty
    if _abi_cache[contract] is not None:
        return

    # load abi
    if contract == 'pool':
        address = '0x8f8ef111b67c04eb1641f5ff19ee54cda062f163'
    elif contract == 'quoter':
        address = quoter
    elif contract == 'tick_lens':
        address = tick_lens
    else:
        raise Exception('unknown contract')

    # get contract abi
    contract_abi = await evm.async_get_contract_abi(contract_address=address)

    # place in cache
    contract_abi_by_name = {
        entry['name']: entry for entry in contract_abi if 'name' in entry
    }
    _abi_cache[contract] = contract_abi_by_name


async def async_get_function_abi(
    function_name: str,
    contract: str,
) -> spec.FunctionABI:
    await _async_load_abi_cache(contract)
    abi_entry = _abi_cache[contract]
    if abi_entry is None:
        raise Exception('must load abi into cache')

    if typing.TYPE_CHECKING:
        return typing.cast(spec.FunctionABI, abi_entry[function_name])
    else:
        return abi_entry[function_name]


async def async_get_event_abi(event_name: str, contract: str) -> spec.EventABI:
    await _async_load_abi_cache(contract)
    abi_entry = _abi_cache[contract]
    if abi_entry is None:
        raise Exception('must load abi into cache')

    if typing.TYPE_CHECKING:
        return typing.cast(spec.EventABI, abi_entry[event_name])
    else:
        return abi_entry[event_name]


factory_event_abis: typing.Mapping[str, spec.EventABI] = {
    'PoolCreated': {
        'anonymous': False,
        'inputs': [
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'token0',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'address',
                'name': 'token1',
                'type': 'address',
            },
            {
                'indexed': True,
                'internalType': 'uint24',
                'name': 'fee',
                'type': 'uint24',
            },
            {
                'indexed': False,
                'internalType': 'int24',
                'name': 'tickSpacing',
                'type': 'int24',
            },
            {
                'indexed': False,
                'internalType': 'address',
                'name': 'pool',
                'type': 'address',
            },
        ],
        'name': 'PoolCreated',
        'type': 'event',
    },
}



from __future__ import annotations

import typing

from ctc import spec

from .. import dex_class_utils
from .. import dex_directory


async def async_get_pool_assets(
    pool: spec.Address,
    *,
    context: spec.Context = None,
    factory: spec.Address | None = None,
) -> typing.Sequence[spec.Address]:
    """get assets of a given DEX pool"""

    from ctc import config

    read_cache, write_cache = config.get_context_cache_read_write(
        schema_name='dex_pools', context=context
    )

    # try obtaining pool from db
    if read_cache:
        from ctc import db

        pool_data = await db.async_query_dex_pool(address=pool, context=context)
        if pool_data is not None:
            assets: typing.MutableSequence[spec.Address] = []
            asset0 = pool_data['asset0']
            if asset0 is not None:
                assets.append(asset0)
            asset1 = pool_data['asset1']
            if asset1 is not None:
                assets.append(asset1)
            asset2 = pool_data['asset2']
            if asset2 is not None:
                assets.append(asset2)
            asset3 = pool_data['asset3']
            if asset3 is not None:
                assets.append(asset3)
            return assets

    # if not in db, acquire using RPC
    if factory is None:
        raise Exception('must specify factory if pool not in db')
    dex_name = dex_directory.get_dex_name_of_factory(
        factory=factory,
        context=context,
    )
    dex = dex_class_utils.get_dex_class(dex_name)
    return await dex.async_get_pool_assets(pool=pool, context=context)




from __future__ import annotations

import typing

from ctc import spec
from .. import rpc_logging
from . import request_utils


@typing.overload
async def async_send(
    request: spec.RpcRequest,
    *,
    context: spec.Context = None,
    raw_output: typing.Literal[True],
    convert_reverts_to: typing.Any = None,
    convert_reverts_to_none: bool = False,
) -> str:
    ...


@typing.overload
async def async_send(
    request: spec.RpcRequest,
    *,
    context: spec.Context = None,
    raw_output: typing.Literal[False] = False,
    convert_reverts_to: typing.Any = None,
    convert_reverts_to_none: bool = False,
) -> spec.RpcResponse:
    ...


@typing.overload
async def async_send(
    request: spec.RpcRequest,
    *,
    context: spec.Context = None,
    raw_output: bool,
    convert_reverts_to: typing.Any = None,
    convert_reverts_to_none: bool = False,
) -> str | spec.RpcResponse:
    ...


async def async_send(
    request: spec.RpcRequest,
    *,
    context: spec.Context = None,
    raw_output: bool = False,
    convert_reverts_to: typing.Any = None,
    convert_reverts_to_none: bool = False,
) -> str | spec.RpcResponse:
    """send RPC request to RPC provider"""

    from ctc import config

    # get provider
    provider = config.get_context_provider(context)
    if provider is None:
        raise Exception('no provider available')

    try:
        logging_rpc_calls = config.get_log_rpc_calls()
    except Exception:
        logging_rpc_calls = False

    if isinstance(request, dict):

        # log request
        if logging_rpc_calls:
            rpc_logging.log_rpc_request(request=request, provider=provider)

        # send request
        raw_response = await async_send_raw(request=request, provider=provider)

        # process response
        output = request_utils._postprocess_response(
            raw_response=raw_response,
            request=request,
            provider=provider,
            raw_output=raw_output,
            convert_reverts_to_none=convert_reverts_to_none,
            convert_reverts_to=convert_reverts_to,
            logging_rpc_calls=logging_rpc_calls,
        )

    elif isinstance(request, list):
        import asyncio

        # circuit break for empty batch request
        if len(request) == 0:
            return []

        # break into individual requests if batch disabled
        if provider.get('disable_batch_requests'):
            coroutines = [
                async_send(
                    subrequest,
                    context=context,
                    raw_output=raw_output,
                    convert_reverts_to=convert_reverts_to,
                    convert_reverts_to_none=convert_reverts_to_none,
                )
                for subrequest in request
            ]
            return await asyncio.gather(*coroutines)

        # chunk request
        request_chunks = request_utils._chunk_request(request=request, provider=provider)

        # log requests
        if logging_rpc_calls:
            for request_chunk in request_chunks:
                rpc_logging.log_rpc_request(
                    request=request_chunk, provider=provider
                )

        # send request chunks
        coroutines = []
        for request_chunk in request_chunks:
            coroutine = async_send_raw(request=request_chunk, provider=provider)
            coroutines.append(coroutine)
        raw_response_chunks = await asyncio.gather(*coroutines)

        # process responses
        output = request_utils._postprocess_plural_response(
            raw_response_chunks=raw_response_chunks,
            request_chunks=request_chunks,
            request=request,
            provider=provider,
            raw_output=raw_output,
            convert_reverts_to_none=convert_reverts_to_none,
            convert_reverts_to=convert_reverts_to,
            logging_rpc_calls=logging_rpc_calls,
        )

    else:
        raise Exception('unknown request type: ' + str(type(request)))

    return output


async def async_send_raw(
    request: spec.RpcRequest,
    provider: spec.Provider,
) -> str:
    """route RPC request to provider according to specified protocol"""

    if provider['protocol'] == 'http':
        from ..rpc_protocols import rpc_http

        return await rpc_http.async_send_http(
            request=request,
            provider=provider,
        )

    elif provider['protocol'] == 'wss':
        from ..rpc_protocols import rpc_websocket

        return await rpc_websocket.async_send_websocket(
            request=request,
            provider=provider,
        )

    else:
        raise Exception(
            'unknown provider protocol: ' + str(provider['protocol'])
        )




"""adapted from py_ecc package

see https://github.com/ethereum/py_ecc/blob/master/py_ecc/secp256k1/secp256k1.py
"""

from __future__ import annotations

import typing


# secp256k1 curve parameters
P = 2**256 - 2**32 - 977
N = 115792089237316195423570985008687907852837564279074904382605163141518161494337
A = 0
B = 7
Gx = 55066263022277343669578718895168534326250603453777594175500187360389116729240
Gy = 32670510020758816978083085130507043184471273380659243275938904335757337482424
G = (Gx, Gy)

secpk1n = 115792089237316195423570985008687907852837564279074904382605163141518161494337


def safe_ord(value: typing.Any) -> int:
    if isinstance(value, int):
        return value
    else:
        return ord(value)


def bytes_to_int(x: bytes) -> int:
    o = 0
    for b in x:
        o = (o << 8) + safe_ord(b)
    return o


def inv(a: int, n: int) -> int:
    if a == 0:
        return 0
    lm, hm = 1, 0
    low, high = a % n, n
    while low > 1:
        r = high // low
        nm, new = hm - lm * r, high - low * r
        lm, low, hm, high = nm, new, lm, low
    return lm % n


def to_jacobian(p: tuple[int, int]) -> tuple[int, int, int]:
    return (p[0], p[1], 1)


def jacobian_double(p: tuple[int, int, int]) -> tuple[int, int, int]:
    if not p[1]:
        return (0, 0, 0)
    ysq = (p[1] ** 2) % P
    S = (4 * p[0] * ysq) % P
    M = (3 * p[0] ** 2 + A * p[2] ** 4) % P
    nx = (M**2 - 2 * S) % P
    ny = (M * (S - nx) - 8 * ysq**2) % P
    nz = (2 * p[1] * p[2]) % P
    return (nx, ny, nz)


def jacobian_add(
    p: tuple[int, int, int], q: tuple[int, int, int]
) -> tuple[int, int, int]:
    if not p[1]:
        return q
    if not q[1]:
        return p
    U1 = (p[0] * q[2] ** 2) % P
    U2 = (q[0] * p[2] ** 2) % P
    S1 = (p[1] * q[2] ** 3) % P
    S2 = (q[1] * p[2] ** 3) % P
    if U1 == U2:
        if S1 != S2:
            return (0, 0, 1)
        return jacobian_double(p)
    H = U2 - U1
    R = S2 - S1
    H2 = (H * H) % P
    H3 = (H * H2) % P
    U1H2 = (U1 * H2) % P
    nx = (R**2 - H3 - 2 * U1H2) % P
    ny = (R * (U1H2 - nx) - S1 * H3) % P
    nz = (H * p[2] * q[2]) % P
    return (nx, ny, nz)


def from_jacobian(p: tuple[int, int, int]) -> tuple[int, int]:
    z = inv(p[2], P)
    return ((p[0] * z**2) % P, (p[1] * z**3) % P)


def multiply(a: tuple[int, int], n: int) -> tuple[int, int]:
    return from_jacobian(jacobian_multiply(to_jacobian(a), n))


def add(a: tuple[int, int], b: tuple[int, int]) -> tuple[int, int]:
    return from_jacobian(jacobian_add(to_jacobian(a), to_jacobian(b)))


# bytes32
def privtopub(privkey: bytes) -> tuple[int, int]:
    return multiply(G, bytes_to_int(privkey))


def deterministic_generate_k(msghash: bytes, priv: bytes) -> int:
    import hmac
    import hashlib

    v = b'\x01' * 32
    k = b'\x00' * 32
    k = hmac.new(k, v + b'\x00' + priv + msghash, hashlib.sha256).digest()
    v = hmac.new(k, v, hashlib.sha256).digest()
    k = hmac.new(k, v + b'\x01' + priv + msghash, hashlib.sha256).digest()
    v = hmac.new(k, v, hashlib.sha256).digest()
    return bytes_to_int(hmac.new(k, v, hashlib.sha256).digest())


# bytes32, bytes32 -> v, r, s (as numbers)
def ecdsa_raw_sign(msghash: bytes, priv: bytes) -> tuple[int, int, int]:

    z = bytes_to_int(msghash)
    k = deterministic_generate_k(msghash, priv)

    r, y = multiply(G, k)
    s = inv(k, N) * (z + r * bytes_to_int(priv)) % N

    v, r, s = (
        27 + ((y % 2) ^ (0 if s * 2 < N else 1)),
        r,
        s if s * 2 < N else N - s,
    )
    return v, r, s


def jacobian_multiply(a: tuple[int, int, int], n: int) -> tuple[int, int, int]:  # type: ignore
    if a[1] == 0 or n == 0:
        return (0, 0, 1)
    if n == 1:
        return a
    if n < 0 or n >= N:
        return jacobian_multiply(a, n % N)
    if (n % 2) == 0:
        return jacobian_double(jacobian_multiply(a, n // 2))
    if (n % 2) == 1:
        return jacobian_add(jacobian_double(jacobian_multiply(a, n // 2)), a)


def ecdsa_raw_recover(
    msghash: bytes,
    vrs: tuple[int, int, int],
) -> tuple[int, int]:
    v, r, s = vrs
    # if v in [0, 1]:
    #     v = v + 27
    if not (27 <= v <= 34):
        raise ValueError('%d must in range 27-31' % v)
    x = r
    xcubedaxb = (x * x * x + A * x + B) % P
    beta = pow(xcubedaxb, (P + 1) // 4, P)
    y = beta if v % 2 ^ beta % 2 else (P - beta)
    # If xcubedaxb is not a quadratic residue, then r cannot be the x coord
    # for a point on the curve, and so the sig is invalid
    if (xcubedaxb - y * y) % P != 0 or not (r % N) or not (s % N):
        raise ValueError(
            "sig is invalid, %d cannot be the x coord for point on curve" % r
        )
    z = bytes_to_int(msghash)
    Gz = jacobian_multiply((Gx, Gy, 1), (N - z) % N)
    XY = jacobian_multiply((x, y, 1), s)
    Qr = jacobian_add(Gz, XY)
    Q = jacobian_multiply(Qr, inv(r, N))
    Q_jacobian = from_jacobian(Q)

    return Q_jacobian



from __future__ import annotations

from .. import query_utils
from . import events_statements


async_query_events = query_utils.wrap_selector_with_connection(
    events_statements.async_select_events,
    'events',
)


async_query_event_queries = query_utils.wrap_selector_with_connection(
    events_statements.async_select_event_queries,
    'events',
)



"""
TODO:
- allow using function selector instead of function name
- allow using quoted json to indicate function abi
- --normalize arg to normalize outputs
- be able to use ERC20 symbols in addresses
"""

from __future__ import annotations

import asyncio
import typing

import toolcli
import toolstr

import ctc.config
from ctc import cli
from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.cli import cli_utils


command_help = """output the result of multiple contract eth_call's

This command can be used in two ways:

[title]1) [option]**Multi-contract Calls**[/option][/title][comment]:[/comment] same call across multiple contracts
    - This is indicated by the `--addresses` parameter
    - Example: `call <function> --addresses <addr1> <addr2> <addr3>`

[title]2) [option]**Multi-block Calls**[/option][/title][comment]:[/comment] same call across multiple blocks
    - This is indicated by either of the `--blocks` or `--time` parameters
    - Example: `call <addr1> <function> --time 30d`

[comment]in each case, `<function>` can be a name, a 4byte selector, or an ABI[/comment]

## Multi-contract Calls

Example use case: Get a wallet's balances for many ERC20 tokens

Can use `--block` to select a specific block for calls

By default will use the first address's function abi for all calls
    [comment](Retrieve separate ABI's for each contract using `--unique-abis`)[/comment]

## Multi-block Calls

Example use case: Get the total supply of an ERC20 over time

Can specify multi-block calls by blocks or by timestamps

Ranges are inclusive of start and end boundaries

### Specifying multi-block calls using `--blocks`

Example block specifications:
    `--blocks 16000 16010 16020 16030`
    `--blocks 16010:16030:10`
    `--blocks 16010:16030 -n 4`
    `--blocks 16030:-20:10`
    `--blocks latest:-20:10`
    [comment](each of these examples would specify the same set of 4 blocks,
     if the latest block were 16030)[/comment]

Can also specify a block range sampled over a regular time interval:
    `--blocks 15000000:16000000:1d` --> samples block range once per day

For block ranges with no interval specified, **all** blocks are used

### Specifying multi-block calls using `--times`

Example time specifications:
    `--times 20220701:20220730 -n 30`
    `--times 20220701:20220730:1d`
    `--times 20220701:29d -n 30`
    `--times 29d:20220701 -n 30`
    [comment](each of these examples specifies the same set of 30 timestamps)[/comment]

Can also use the current time as an implicit default:
    `--times 100d::1d`
    `--times 100d -n 101`
    `--times 100d: -n 101`
    `--times 100d:: -n 101`
    [comment](each of these examples specifies the same set of 100 timestamps)[/comment]

Can indicate timestamps using a variety of formats:

[timestamp table]

See the `tooltime` package for more information about time specifications

A default of `-n 25` is used if no `-n` is specified for time range

## Outputs

A table of outputs is displayed by default, omit with `--no-table`

Numerical outputs of multi-block calls are charted, omit with `--no-chart`

Use `--export` to output data to a json or csv file
More data is included when exporting to a file
"""


def get_command_help(parse_spec: toolcli.ParseSpec) -> str:
    styles = cli.get_cli_styles()
    stylized = stylize_markdown(command_help, styles=styles)
    stylized = stylized.replace(
        '[timestamp table]', get_timestamp_format_table()
    )
    stylized = stylized.replace('[comment]', '[' + styles['comment'] + ']')
    stylized = stylized.replace('[/comment]', '[/' + styles['comment'] + ']')
    return stylized


def get_timestamp_format_table() -> str:
    rows = [
        [
            'number + letter\n[comment](one of {yMdhms})[/comment]',
            'timelength',
            '100d\n24h',
        ],
        ['YYYY', 'start of year', '2018\n2022'],
        ['YYYYMM\nYYYY-MM', 'start of month', '201802\n2022-06'],
        ['YYYYMMDD\nYYYY-MM-DD', 'start of day', '20180228\n2022-06-01'],
        [
            'YYYYMMDD_HHMMSS\nYYYY-MM-DD_HH:MM',
            'minute in day',
            '20180228\n2022-06-01_04:30',
        ],
        [
            'YYYYMMDD_HHMMSS\nYYYY-MM-DD_HH:MM:SS',
            'second in day',
            '20180228\n2022-06-01_04:30:40',
        ],
        ['large number', 'unix timestamp', '1600000000\n1700000000'],
    ]

    labels = [
        'timestamp formats',
        'meaning',
        'examples',
    ]

    styles = cli.get_cli_styles()

    as_str = toolstr.print_multiline_table(
        rows,
        labels=labels,
        return_str=True,
        label_style=styles['title'],
        border=styles['comment'],
        column_styles={
            'timestamp formats': styles['option'],
            'meaning': styles['description'],
            'examples': styles['option'],
        },
        indent=4,
    )

    if not isinstance(as_str, str):
        raise Exception('did not return str')

    return as_str


def stylize_markdown(text: str, styles: toolcli.StyleTheme) -> str:
    import re

    lines = text.split('\n')

    code_regex = re.compile('`.*?`')
    bold_regex = re.compile('\*\*.*?\*\*')

    styled_lines = []
    for i, line in enumerate(lines):
        # stylize titles
        if line.startswith('#'):
            line = toolstr.add_style(line, styles['title'])

        if line.startswith('    - '):
            line = toolstr.add_style(line[:6], styles['title']) + line[6:]

        # replace code snippets
        for match in code_regex.findall(line):
            styled_match = toolstr.add_style(match.strip('`'), styles['option'])
            line = line.replace(match, styled_match)

        for match in bold_regex.findall(line):
            styled_match = toolstr.add_style(match.strip('*'), 'bold')
            line = line.replace(match, styled_match)

        styled_lines.append(line)

    return '\n'.join(styled_lines)


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_calls_command,
        'help': get_command_help,
        'args': [
            {
                'name': 'address_and_or_function',
                'nargs': '*',
                'help': '<see above>',
            },
            #
            # multi-block queries
            {
                'name': ['-b', '--blocks'],
                'nargs': '+',
                'help': 'block list or block range (see above)',
            },
            {
                'name': ['-t', '--times'],
                'nargs': '+',
                'help': 'timestamp list or time range (see above)',
            },
            {
                'name': '-n',
                'help': 'number of calls to make in given range',
                'type': int,
            },
            #
            # multi-address queries
            {
                'name': ['-a', '--addresses'],
                'nargs': '+',
                'help': 'addresses to point calls toward',
            },
            {'name': '--block', 'help': 'block number for calls'},
            {
                'name': '--unique-abis',
                'help': 'retrieve separate ABI for each address',
            },
            #
            # output parameters
            {
                'name': '--export',
                'default': 'stdout',
                'help': 'file path for output (.json or .csv)',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {
                'name': '--no-table',
                'action': 'store_true',
                'help': 'do not output a table',
            },
            {
                'name': '--no-chart',
                'action': 'store_true',
                'help': 'do not output a chart',
            },
            #
            # advanced parameters
            {
                'name': '--from',
                'help': 'address that calls should come from',
                'dest': 'from_address',
            },
            {
                'name': '--normalize',
                'help': 'normalize output by value, default is 1) erc20 decimals or 2) "1e18"',
                'nargs': '*',
            },
            {
                'name': '--blocks-gte-times',
                'help': 'find the blocks >= timestamps instead of <=',
                'action': 'store_true',
            },
        ],
        'examples': {
            '<function> [<function_parameters>] --addresses <addresses>': {
                'description': '1) multi-contract template: same call across multiple addresses',
                'runnable': False,
            },
            '<address> <function> [<function_parameters>] --blocks <blocks>': {
                'description': '2) multi-block template: same call across multiple blocks',
                'runnable': False,
            },
            'balanceOf 0xbebc44782c7db0a1a60cb6fe97d0b483032ff1c7 --addresses 0x6b175474e89094c44da98b954eedeac495271d0f 0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48 0xdac17f958d2ee523a2206206994597c13d831ec7 --normalize': {
                'description': 'Balance of DAI, USDC, and USDT in Curve 3pool',
                'runnable': True,
            },
            '0x6b175474e89094c44da98b954eedeac495271d0f balanceOf 0xbebc44782c7db0a1a60cb6fe97d0b483032ff1c7 --times 2021:2022:1w --normalize': {
                'description': 'Weekly balance of DAI in Curve 3pool throughout year 2021',
                'runnable': True,
            },
            '0x6b175474e89094c44da98b954eedeac495271d0f balanceOf 0xbebc44782c7db0a1a60cb6fe97d0b483032ff1c7 --blocks latest:-1000000 -n 20 --normalize': {
                'description': 'Balance of DAI in Curve 3pool over the past 1 million blocks',
                'runnable': True,
            },
        },
    }


async def async_calls_command(
    *,
    address_and_or_function: typing.Sequence[str],
    blocks: typing.Sequence[str] | None,
    times: typing.Sequence[str] | None,
    n: int | None,
    addresses: typing.Sequence[str] | None,
    unique_abis: bool,
    block: str,
    export: str,
    overwrite: bool,
    no_table: bool,
    no_chart: bool,
    from_address: spec.Address | None,
    normalize: typing.Sequence[str],
    blocks_gte_times: bool,
) -> None:
    # check whether using multi-contract query or multi-block query
    multi_contract_query = addresses is not None
    multi_block_query = blocks is not None or times is not None
    if multi_contract_query and multi_block_query:
        raise Exception('cannot specify --addresses and --blocks or --times')

    # resolve from address
    if from_address is not None:
        from_address = await evm.async_resolve_address(
            from_address, block=block
        )

    if multi_contract_query:
        df: spec.DataFrame | None = await async_perform_multi_contract_call(
            address_and_or_function=address_and_or_function,
            addresses=addresses,
            block=block,
            unique_abis=unique_abis,
            from_address=from_address,
            normalize=normalize,
            export=export,
        )

    elif multi_block_query:
        df = await async_perform_multi_block_call(
            address_and_or_function=address_and_or_function,
            blocks=blocks,
            times=times,
            n=n,
            addresses=addresses,
            block=block,
            export=export,
            no_table=no_table,
            no_chart=no_chart,
            from_address=from_address,
            normalize=normalize,
            blocks_gte_times=blocks_gte_times,
        )

    else:
        raise Exception('must specify either --blocks or --addresses')

    if export != 'stdout':
        if df is None:
            raise Exception('did not obtain dataframe of data')
        cli_utils.output_data(data=df, output=export, overwrite=overwrite)


async def async_perform_multi_contract_call(
    *,
    address_and_or_function: typing.Sequence[str],
    addresses: typing.Sequence[spec.Address] | None,
    block: str,
    unique_abis: bool,
    from_address: spec.Address | None,
    normalize: typing.Sequence[str],
    export: str,
) -> spec.DataFrame | None:
    if addresses is not None:
        addresses = await evm.async_resolve_addresses(addresses, block=block)
    if addresses is None:
        raise Exception('could not detect addresses')

    # multi-contract queries
    if block is None:
        block = 'latest'

    function, *function_parameters = address_and_or_function

    # assert that all address functions have the same number of outputs
    function_abi = await evm.async_parse_function_str_abi(
        function,
        contract_address=addresses[0],
    )
    n_outputs = len(function_abi['outputs'])
    for to_address in addresses[1:]:
        other_function_abi = await evm.async_get_function_abi(
            contract_address=to_address,
            function_name=function_abi['name'],
        )
        if len(other_function_abi['inputs']) != len(function_abi['inputs']):
            print('not all addresses have same number of function inputs')
        if len(other_function_abi['outputs']) != n_outputs:
            print('not all addresses have same number of function outputs')

    # fetch data
    if unique_abis:
        import asyncio

        coroutines = [
            rpc.async_batch_eth_call(
                to_address=address,
                function_abi=function_abi,
                function_parameters=function_parameters,
                block_number=block,
                from_address=from_address,
            )
            for address in addresses
        ]
        raw_results = await asyncio.gather(*coroutines)
    else:
        raw_results = await rpc.async_batch_eth_call(
            to_addresses=addresses,
            function_abi=function_abi,
            function_parameters=function_parameters,
            block_number=block,
            from_address=from_address,
        )

    # name based on first contract's abi
    output_names = evm.get_function_output_names(
        function_abi, human_readable=True
    )

    if normalize is not None:
        if len(normalize) == 0:
            if len(output_names) != 1:
                raise Exception(
                    'can only normalize functions that have a single output'
                )
            try:
                results: typing.Sequence[
                    typing.Any
                ] = await evm.async_normalize_erc20s_quantities(
                    quantities=raw_results,
                    tokens=addresses,
                )
            except Exception:
                results = [result / 1e18 for result in raw_results]
        elif len(normalize) == 1:
            factor = float(normalize[0])
            results = [subresult / factor for subresult in raw_results]
        else:
            raise Exception('--normalize should have at most 1 argument')
    else:
        results = raw_results

    styles = cli.get_cli_styles()
    toolstr.print_text_box('Multi-contract calls', style=styles['title'])
    if block is None:
        block = 'latest'
    cli.print_bullet(key='block', value=block)
    cli.print_bullet(key='function', value=function_abi.get('name', '\[none]'))
    cli.print_bullet(key='function_args', value='')
    for fp, function_parameter in enumerate(function_parameters):
        cli.print_bullet(value=function_parameter, number=fp + 1, indent=4)
    print()

    if export == 'stdout':
        rows = []
        for r, result in enumerate(results):
            row: list[typing.Any] = [addresses[r]]
            if len(output_names) == 1:
                if normalize is not None:
                    row.append(result)
                else:
                    row.append(str(result))
            else:
                for subresult in result:
                    row.append(str(subresult))
            rows.append(row)
        labels: list[typing.Any] = ['address']
        for output_name in output_names:
            labels.append(output_name)
        toolstr.print_table(
            rows,
            labels=labels,
            border=styles['comment'],
            label_style=styles['title'],
            column_styles=[styles['metavar']]
            + [styles['description']] * len(output_names),
            column_formats={
                'balanceOf': {'trailing_zeros': True, 'decimals': 2},
            },
        )
        return None

    else:
        import polars as pl

        names = [name if name is not None else '' for name in output_names]
        df = pl.DataFrame(results, names)
        df.with_columns(pl.Series('to_address', addresses))

        return df


async def async_perform_multi_block_call(
    *,
    address_and_or_function: typing.Sequence[str],
    blocks: typing.Sequence[str] | None,
    times: typing.Sequence[str] | None,
    n: int | None,
    addresses: typing.Sequence[str] | None,
    block: str,
    export: str,
    no_table: bool,
    no_chart: bool,
    from_address: spec.Address | None,
    normalize: typing.Sequence[str],
    blocks_gte_times: bool,
) -> spec.DataFrame | None:
    # historical queries
    if block is not None:
        raise Exception(
            'cannot specify both --block alongside --blocks or --times'
        )
    if addresses is not None:
        raise Exception('cannot specify --addresses in multi block call')

    # parse address, function, and function parameters
    (
        to_address,
        function,
        *function_parameters,
    ) = address_and_or_function
    to_address = await evm.async_resolve_address(to_address, block=block)
    function_abi = await evm.async_parse_function_str_abi(
        function,
        contract_address=to_address,
    )

    # parse blocks or times
    if blocks is not None:
        # parse blocks
        block_numbers = await cli_utils.async_parse_block_slice(blocks, n=n)

    elif times is not None:
        # parse timestamps
        import tooltime

        if len(times) != 1:
            raise Exception('--times specified improperly')

        raw_timestamps = tooltime.parse_timeslice(times[0], n=n)
        timestamps = [int(timestamp) for timestamp in raw_timestamps]
        if blocks_gte_times:
            mode = '>='
        else:
            mode = '<='
        block_numbers = await evm.async_get_blocks_of_timestamps(
            timestamps,
            mode=mode,
            # mode='>=',
        )

    else:
        raise Exception('must specify --blocks or --times')

    if not no_table:
        block_timestamps_coroutine = evm.async_get_block_timestamps(
            block_numbers
        )
        block_timestamps_task = asyncio.create_task(block_timestamps_coroutine)

    # fetch data
    results = await rpc.async_batch_eth_call(
        to_address=to_address,
        function_abi=function_abi,
        function_parameters=function_parameters,
        block_numbers=block_numbers,
        from_address=from_address,
    )

    output_names = evm.get_function_output_names(
        function_abi, human_readable=True
    )
    if normalize is not None:
        if len(normalize) == 0:
            if len(output_names) != 1:
                raise Exception(
                    'can only normalize functions that have a single output'
                )
            try:
                results = await evm.async_normalize_erc20_quantities(
                    quantities=results,
                    token=to_address,
                )
            except Exception:
                result = [result / 1e18 for result in results]
        elif len(normalize) == 1:
            factor = float(normalize[0])
            results = [subresult / factor for subresult in results]
        else:
            raise Exception('--normalize should have at most 1 argument')

    styles = cli.get_cli_styles()
    toolstr.print_text_box('Multi-block calls', style=styles['title'])
    cli.print_bullet(
        key='address', value=toolstr.add_style(to_address, styles['metavar'])
    )
    cli.print_bullet(key='function', value=function_abi.get('name', '\[none]'))
    cli.print_bullet(key='function_args', value='')
    for fp, function_parameter in enumerate(function_parameters):
        cli.print_bullet(value=function_parameter, number=fp + 1, indent=4)
    cli.print_bullet(key='# blocks', value=str(len(block_numbers)))
    print()

    if export == 'stdout':
        import tooltime

        if not no_table:
            block_timestamps = await block_timestamps_task

            rows = []
            for r, result in enumerate(results):
                row = []
                row.append(block_numbers[r])
                age = tooltime.get_age(block_timestamps[r], 'TimelengthPhrase')
                age = ', '.join(age.split(', ')[:2])
                row.append(age)
                if len(output_names) == 1:
                    row.append(result)
                else:
                    for subresult in result:
                        row.append(subresult)

                rows.append(row)
            labels = [
                'block',
                'block age',
            ]
            labels += output_names
            column_styles = {
                'block': styles['option'],
                'block age': styles['option'],
            }
            column_formats = {
                'block': {'commas': False},
            }
            for output_name in output_names:
                column_styles[output_name] = styles['description']
                if normalize is not None:
                    column_formats[output_name] = {'trailing_zeros': True}
            print()
            toolstr.print_table(
                rows,
                labels=labels,
                border=styles['comment'],
                label_style=styles['title'],
                column_styles=column_styles,
                column_formats=column_formats,
                indent=4,
            )

        if not no_chart:
            xvals = block_numbers
            yvals = results
            plot = toolstr.render_line_plot(
                xvals=xvals,
                yvals=yvals,
                n_rows=10,
                n_columns=60,
                line_style=styles['description'],
                chrome_style=styles['comment'],
                tick_label_style=styles['metavar'],
                xaxis_kwargs={'tick_label_format': None, 'n_ticks': 2},
                char_dict=ctc.config.get_cli_chart_charset(),
            )
            print()
            print()
            toolstr.print(
                toolstr.hjustify(output_names[0], 'center', 70),
                indent=4,
                style=styles['title'],
            )
            toolstr.print(plot, indent=4)

        return None

    else:
        import polars as pl

        # format into dataframe
        df = pl.DataFrame(results, output_names)
        df = df.insert_at_idx(0, pl.Series('block', block_numbers))
        return df




from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

from .. import irm_metadata
from .. import token_metadata
from ... import rari_abis


async def async_get_supply_interest_by_block(
    ctoken: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    fill_empty: bool = False,
    empty_token: typing.Any = None,
) -> int | float | None:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['supplyRatePerBlock'],
        fill_empty=fill_empty,
        empty_token=empty_token,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    output: int | float = result
    if fill_empty and output is None:
        return None
    if normalize:
        output /= 1e18
    return output


async def async_get_borrow_interest_by_block(
    ctoken: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    fill_empty: bool = False,
    empty_token: typing.Any = None,
) -> int | float | None:
    result = await rpc.async_eth_call(
        to_address=ctoken,
        block_number=block,
        function_abi=rari_abis.ctoken_function_abis['borrowRatePerBlock'],
        fill_empty=fill_empty,
        empty_token=empty_token,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    output: int | float = result
    if fill_empty and output is None:
        return None
    if normalize:
        output /= 1e18
    return output


async def async_get_supply_apy(
    ctoken: spec.Address,
    *,
    blocks_per_year: int | None = None,
    block: spec.BlockNumberReference = 'latest',
    fill_empty: bool = False,
    empty_token: typing.Any = None,
) -> float | None:
    supply_interest_per_block = await async_get_supply_interest_by_block(
        ctoken=ctoken,
        block=block,
        normalize=True,
        fill_empty=fill_empty,
        empty_token=empty_token,
    )
    if supply_interest_per_block is None:
        if fill_empty:
            return None
        else:
            raise Exception('could not determine supply interest per block')
    if blocks_per_year is None:
        irm = await token_metadata.async_get_ctoken_irm(ctoken, block=block)
        blocks_per_year = await irm_metadata.async_get_irm_blocks_per_year(
            irm,
            block=block,
        )

    try:
        return (1 + supply_interest_per_block) ** blocks_per_year - 1
    except OverflowError:
        return float('inf')


async def async_get_borrow_apy(
    ctoken: spec.Address,
    *,
    blocks_per_year: int | None = None,
    block: spec.BlockNumberReference = 'latest',
    fill_empty: bool = False,
    empty_token: typing.Any = None,
) -> float | None:

    borrow_interest_per_block = await async_get_borrow_interest_by_block(
        ctoken=ctoken,
        block=block,
        normalize=True,
        fill_empty=fill_empty,
        empty_token=empty_token,
    )
    if borrow_interest_per_block is None:
        if fill_empty:
            return None
        else:
            raise Exception('could not determine supply interest per block')

    if blocks_per_year is None:
        irm = await token_metadata.async_get_ctoken_irm(ctoken, block=block)
        blocks_per_year = await irm_metadata.async_get_irm_blocks_per_year(
            irm,
            block=block,
        )

    try:
        return (1 + borrow_interest_per_block) ** blocks_per_year - 1
    except OverflowError:
        return float('inf')



from __future__ import annotations

import toolcli

from ctc import evm
from ctc import spec
from ctc.protocols import ens_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_reverse_command,
        'help': 'reverse ENS lookup address',
        'args': [
            {'name': 'address', 'help': 'address of reverse lookup'},
            {'name': '--block', 'help': 'block number'},
        ],
        'examples': [
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045',
            '0xd8da6bf26964af9d7eed9e03e53415d37aa96045 --block 14000000',
        ],
    }


async def async_reverse_command(
    address: spec.Address,
    block: spec.BlockNumberReference,
) -> None:
    if block is not None:
        block = evm.standardize_block_number(block)
    name = await ens_utils.async_reverse_lookup(address, block=block)
    print(name)



from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec

from .. import rari_abis
from . import directory_metadata
from . import token_metadata


async def async_get_pool_ctoken(
    comptroller: spec.Address,
    underlying: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
) -> spec.Address:
    result = await rpc.async_eth_call(
        to_address=comptroller,
        block_number=block,
        function_abi=rari_abis.comptroller_function_abis['cTokensByUnderlying'],
        function_parameters=[underlying],
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_ctokens(
    comptroller: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
) -> tuple[spec.Address]:
    result = await rpc.async_eth_call(
        to_address=comptroller,
        block_number=block,
        function_abi=rari_abis.comptroller_function_abis['getAllMarkets'],
    )
    if not isinstance(result, tuple) or not all(
        isinstance(item, str) for item in result
    ):
        raise Exception('invalid rpc result')
    return typing.cast(typing.Tuple[str], result)


async def async_get_pool_underlying_tokens(
    *,
    ctokens: typing.Sequence[spec.Address] | None = None,
    comptroller: spec.Address | None = None,
    block: spec.BlockNumberReference = 'latest',
) -> typing.Mapping[spec.Address, spec.Address]:
    import asyncio

    if ctokens is None:
        if comptroller is None:
            raise Exception('specify comptroller')
        ctokens = await async_get_pool_ctokens(comptroller, block=block)

    coroutines = [
        token_metadata.async_get_ctoken_underlying(ctoken=ctoken)
        for ctoken in ctokens
    ]
    underlyings = await asyncio.gather(*coroutines)
    return dict(zip(ctokens, underlyings))


async def async_get_pool_oracle(
    comptroller: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
) -> spec.Address:
    result = await rpc.async_eth_call(
        to_address=comptroller,
        block_number=block,
        function_abi=rari_abis.comptroller_function_abis['oracle'],
    )
    if not isinstance(result, str):
        raise Exception('invalid rpc result')
    return result


async def async_get_pool_name(
    comptroller: spec.Address,
    *,
    all_pools: typing.Sequence[typing.Sequence[typing.Any]] | None = None,
    block: spec.BlockNumberReference = 'latest',
) -> str:
    comptroller = comptroller.lower()
    if all_pools is None:
        all_pools = await directory_metadata.async_get_all_pools(block=block)
    for pool in all_pools:
        if pool[2] == comptroller:
            output = pool[0]
            if not isinstance(output, str):
                raise Exception('invalid rpc result')
            return output
    else:
        raise Exception('could not find pool')



from __future__ import annotations

from ctc import spec
from .. import abi_utils


async def async_is_erc721(
    contract_address: spec.Address | None,
    *,
    contract_abi: spec.ContractABI | None = None,
    context: spec.Context = None,
) -> bool:
    """return whether contract at address implements erc721 standard"""

    if contract_abi is None:
        if contract_address is None:
            raise Exception('specify either contract_address or contract_abi')
        contract_abi = await abi_utils.async_get_contract_abi(
            contract_address, context=context
        )

    return _is_erc721_contract_abi(contract_abi)


def _is_erc721_contract_abi(contract_abi: spec.ContractABI) -> bool:
    """return whether contract abi implements erc721 standard"""

    erc721_signatures = {
        #
        # function signatures
        'balanceOf(address)',
        'ownerOf(uint256)',
        'safeTransferFrom(address,address,uint256,bytes)',
        'safeTransferFrom(address,address,uint256)',
        'transferFrom(address,address,uint256)',
        'approve(address,uint256)',
        'setApprovalForAll(address,bool)',
        'getApproved(uint256)',
        'isApprovedForAll(address,address)',
        #
        # event signatures
        'Transfer(address,address,uint256)',
        'Approval(address,address,uint256)',
        'ApprovalForAll(address,address,bool)',
    }

    contract_signatures = set()
    for item in contract_abi:
        if item.get('type') == 'function':
            signature = abi_utils.get_function_signature(function_abi=item)  # type: ignore
            contract_signatures.add(signature)
        elif item.get('type') == 'event':
            signature = abi_utils.get_event_signature(event_abi=item)  # type: ignore
            contract_signatures.add(signature)

    missing_signatures = erc721_signatures - contract_signatures

    return len(missing_signatures) == 0




from __future__ import annotations

import toolsql

dex_pools_schema: toolsql.DBSchemaShorthand = {
    'tables': {
        'dex_pools': {
            'columns': [
                {'name': 'address', 'type': 'Text', 'primary': True},
                {'name': 'factory', 'type': 'Text', 'index': True},
                {'name': 'asset0', 'type': 'Text', 'index': True},
                {'name': 'asset1', 'type': 'Text', 'index': True},
                {'name': 'asset2', 'type': 'Text', 'index': True},
                {'name': 'asset3', 'type': 'Text', 'index': True},
                {'name': 'creation_block', 'type': 'Integer', 'index': True},
                {'name': 'fee', 'type': 'Integer'},
                {'name': 'additional_data', 'type': 'JSON'},
                {'name': 'priority', 'type': 'Integer', 'index': True},
            ],
        },
        'dex_pool_factory_queries': {
            'columns': [
                {'name': 'factory', 'type': 'Text', 'primary': True},
                {'name': 'last_scanned_block', 'type': 'Integer'},
            ],
        },
    },
}



from __future__ import annotations

import asyncio
import typing

import toolcli
import toolstr

from ctc import cli
from ctc.cli import cli_utils
from ctc import evm
from ctc import spec
from ctc.defi import dex_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_dex_pools_command,
        'help': 'list dex pools',
        'args': [
            {
                'name': 'tokens',
                'help': 'token symbols or addresses (separate using spaces)',
                'nargs': '*',
            },
            {
                'name': '--update',
                'help': 'run update to index all recently created pools',
                'action': 'store_true',
            },
            {
                'name': '--created',
                'help': 'specify start and/or end of when pool was created',
            },
            {
                'name': '--dex',
                'help': 'name of dex (e.g. balancer, curve, uniswap-v2)',
            },
            {
                'name': '--factory',
                'help': 'address of pool factory',
            },
            {
                'name': '--all-pools',
                'help': 'display all pools instead of just the first 1000',
                'action': 'store_true',
            },
            {
                'name': '--compact',
                'help': 'use compact view for higher information density',
                'action': 'store_true',
            },
            {
                'name': ['--verbose', '-v'],
                'help': 'show additional data',
                'action': 'store_true',
            },
            {
                'name': '--json',
                'help': 'output data as json',
                'dest': 'json_output',
                'action': 'store_true',
            },
            {
                'name': '--csv',
                'help': 'output data as json',
                'dest': 'csv_output',
                'action': 'store_true',
            },
            {
                'name': '--export',
                'help': 'file path to save output to',
            },
            {
                'name': '--overwrite',
                'action': 'store_true',
                'help': 'specify that output path can be overwritten',
            },
            {
                'name': '--sort',
                'help': 'column to sort pools by',
                'nargs': '+',
            },
        ],
        'examples': [
            'CRV',
            'DAI --dex balancer',
            'DAI --created 14000000:15000000',
        ],
    }


def simplify_name(name: str) -> str:
    return name.lower().replace('-', '').replace(' ', '').replace('_', '')


async def async_dex_pools_command(
    *,
    tokens: typing.Sequence[spec.Address | str],
    dex: spec.Address | str | None,
    update: bool,
    created: str | None,
    factory: spec.Address | None,
    all_pools: bool,
    compact: bool,
    verbose: bool,
    json_output: bool,
    csv_output: bool,
    export: str | None,
    overwrite: bool,
    sort: typing.Sequence[str] | None,
) -> None:

    factory_dexes = dex_utils.get_dex_names_of_factories(network='ethereum')

    coroutines = [evm.async_get_erc20_address(token) for token in tokens]
    assets = await asyncio.gather(*coroutines)

    if created is not None:
        start_block, end_block = await cli_utils.async_parse_block_range(
            created
        )
    else:
        start_block = None
        end_block = None
    dex_pools = await dex_utils.async_get_pools(
        assets=assets,
        factory=factory,
        dex=dex,
        start_block=start_block,
        end_block=end_block,
        update=update,
    )

    # alternative output formats
    if json_output and export is None:
        import json

        as_str = json.dumps(dex_pools)
        print(as_str)
        return
    if csv_output and export is None:
        import polars as pl

        df = pl.DataFrame(dex_pools)
        csv_str = df.write_csv(None)
        print(csv_str)
        return
    if export:
        import polars as pl

        df = pl.DataFrame(dex_pools)
        cli_utils.output_data(df, output=export, overwrite=overwrite)
        print('saved output to', export)
        return

    styles = cli.get_cli_styles()

    toolstr.print(
        toolstr.add_style(str(len(dex_pools)), styles['description'] + ' bold'),
        'pools found',
    )
    print()

    max_pools = 1000
    if len(dex_pools) > max_pools and not all_pools:
        clipped = len(dex_pools) - max_pools
        dex_pools = dex_pools[:max_pools]
    else:
        clipped = 0

    all_assets: set[str | None] = set()
    n_assets = 2
    for dex_pool in dex_pools:
        all_assets.add(dex_pool['asset0'])
        all_assets.add(dex_pool['asset1'])

        asset2 = dex_pool['asset2']
        if asset2 is not None:
            all_assets.add(dex_pool['asset2'])
            n_assets = 3

        asset3 = dex_pool['asset3']
        if asset3 is not None:
            all_assets.add(dex_pool['asset3'])
            n_assets = 4

    ordered_assets = [asset for asset in all_assets if asset is not None]
    symbols = await evm.async_get_erc20s_symbols(
        ordered_assets,
        context=dict(
            provider={'chunk_size': 10, 'convert_reverts_to_none': True}
        ),
    )

    symbols = [symbol.replace('\x00', '') for symbol in symbols]

    for i in range(len(ordered_assets)):
        if ordered_assets[i] == '0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee':
            symbols[i] = 'ETH'
    assets_to_symbols: typing.Mapping[str | None, str] = dict(
        zip(ordered_assets, symbols)
    )

    # get block times
    if verbose:
        import tooltime

        all_blocks = list(
            {dex_pool['creation_block'] for dex_pool in dex_pools}
        )
        block_timestamps = await evm.async_get_block_timestamps(all_blocks)
        timestamps_of_blocks = dict(zip(all_blocks, block_timestamps))

    rows = []
    for dex_pool in dex_pools:
        row: list[str | None] = [
            dex_pool['address'],
            factory_dexes.get(dex_pool['factory']),
            assets_to_symbols.get(dex_pool['asset0']),
            assets_to_symbols.get(dex_pool['asset1']),
        ]
        if n_assets >= 3:
            row.append(assets_to_symbols.get(dex_pool['asset2'], ''))
        if n_assets >= 4:
            row.append(assets_to_symbols.get(dex_pool['asset3'], ''))

        if verbose:
            block = dex_pool['creation_block']
            timestamp = timestamps_of_blocks[block]
            age = tooltime.get_age(timestamp, 'TimelengthPhrase')
            age = ' '.join(age.split(' ')[:2]).strip(',')

            row.append(str(block))
            row.append(age)

        rows.append(row)

    labels = [
        'address',
        'dex',
        'asset0',
        'asset1',
    ]
    if n_assets >= 3:
        labels.append('asset2')
    if n_assets >= 4:
        labels.append('asset3')
    if verbose:
        labels.append('creation\nblock')
        labels.append('age')

    if sort is not None:
        sort_indices = [labels.index(column) for column in sort]
    else:
        sort_indices = [
            labels.index('dex'),
            labels.index('asset0'),
            labels.index('asset1'),
        ]
    rows = sorted(
        rows,
        key=lambda row: tuple(
            row[index] if row[index] else '' for index in sort_indices
        ),
    )

    if not verbose:
        max_column_widths: typing.Mapping[str | int, int] | None = {
            'asset0': 10,
            'asset1': 10,
            'asset2': 10,
            'asset3': 10,
        }
    else:
        max_column_widths = None

    toolstr.print_table(
        rows,
        labels=labels,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'address': styles['metavar'],
            'asset0': styles['description'] + ' bold',
            'asset1': styles['description'] + ' bold',
            'asset2': styles['description'] + ' bold',
            'asset3': styles['description'] + ' bold',
        },
        max_table_width=toolcli.get_n_terminal_cols(),
        max_column_widths=max_column_widths,
        compact=compact,
    )

    if clipped:
        print()
        toolstr.print(
            'showing only first',
            max_pools,
            'pools, use',
            toolstr.add_style('--all', styles['option']),
            'to display all',
            clipped + max_pools,
            'pools',
            style=styles['comment'],
        )




from __future__ import annotations

import json
import os
import typing

import toolcli
import toolstr

import ctc
import ctc.config
from ctc import spec
from ctc.cli import cli_utils
from .. import config_spec
from .. import config_read
from .. import upgrade_utils


def load_old_config(
    convert_to_latest: bool,
) -> typing.Mapping[typing.Any, typing.Any]:

    # get old config path
    config_path: str | None = os.environ.get(config_spec.config_path_env_var)
    if config_path in ['', None]:
        config_path = config_spec.default_config_path

    # load old config file
    if isinstance(config_path, str) and os.path.isfile(config_path):
        with open(config_path) as f:
            old_config: typing.Mapping[typing.Any, typing.Any] = json.load(f)
    else:
        old_config = {}

    old_providers = old_config.get('providers')
    if old_providers is not None:
        for provider_name, provider in list(old_providers.items()):
            for provider_key in spec.provider_keys:
                if provider_key not in provider:
                    if provider_key in spec.default_provider_settings:
                        provider.setdefault(
                            provider_key,
                            spec.default_provider_settings[provider_key],
                        )
                    else:
                        print('skipping provider, missing essential keys')
                        del old_providers[provider_name]

    # upgrade config file if need be
    config_version = old_config.get('config_spec_version')
    if config_version is not None:
        config_stable_version = upgrade_utils.get_stable_version(config_version)
    else:
        config_stable_version = None
    ctc_stable_version = upgrade_utils.get_stable_version(ctc.__version__)
    if convert_to_latest and config_stable_version != ctc_stable_version:
        try:
            old_config = upgrade_utils.upgrade_config(old_config)
        except spec.ConfigUpgradeError:
            print()
            print('old config could not be processed, skipping it')
            old_config = {}

    return old_config


def setup_config_path() -> None:

    # create parent dir
    config_path = config_read.get_config_path(raise_if_dne=False)
    parent_dir = os.path.dirname(config_path)
    os.makedirs(parent_dir, exist_ok=True)


def write_new_config(
    *,
    network_data: spec.PartialConfig,
    db_data: spec.PartialConfig,
    data_dir_data: spec.PartialConfig,
    cli_data: spec.PartialConfig,
    styles: toolcli.StyleTheme,
    overwrite: bool = False,
    headless: bool = False,
) -> None:

    import json

    config_path = config_read.get_config_path(raise_if_dne=False)

    networks = {
        str(key): value for key, value in network_data['networks'].items()
    }
    default_providers = {
        str(key): value
        for key, value in network_data['default_providers'].items()
    }

    config: spec.JsonConfig = {
        'config_spec_version': ctc.__version__,
        'data_dir': data_dir_data['data_dir'],
        'networks': networks,
        'providers': network_data['providers'],
        'default_network': network_data['default_network'],
        'default_providers': default_providers,
        'db_configs': db_data['db_configs'],
        'log_rpc_calls': data_dir_data['log_rpc_calls'],
        'log_sql_queries': data_dir_data['log_sql_queries'],
        'cli_color_theme': cli_data['cli_color_theme'],
        'cli_chart_charset': cli_data['cli_chart_charset'],
        'context_cache_rules': [
            {
                'backend': 'main',
                'read': True,
                'write': True,
            }
        ],
    }
    print()
    print()
    toolstr.print('## Creating Configuration File', style=styles['title'])
    if os.path.isfile(config_path):
        with open(config_path, 'r') as f:
            old_config_raw = json.load(f)
        write_new = json.dumps(config, sort_keys=True) != json.dumps(
            old_config_raw, sort_keys=True
        )

        # print updated keys
        if write_new:

            # detect updated keys
            changed_keys: set[str] = set()
            for key, value in config.items():
                if json.dumps(value, sort_keys=True) != json.dumps(
                    old_config_raw.get(key), sort_keys=True
                ):
                    changed_keys.add(key)
            for key in old_config_raw.keys():
                if key not in config:
                    changed_keys.add(key)

            print()
            if changed_keys == set(config.keys()):
                print('All keys updated')
            else:
                print('Updated keys:')
                for key in sorted(changed_keys):
                    cli_utils.print_bullet(
                        value=key, key_style=styles['description']
                    )
                    old_value = old_config_raw.get(key)
                    new_value = config.get(key)
                    if isinstance(old_value, dict) and isinstance(
                        new_value, dict
                    ):
                        changed_values = {}
                        for k, v in new_value.items():
                            if v != old_value.get(k):
                                changed_values[k] = (old_value.get(k), v)
                        if len(changed_values) > 0:
                            for k, (old_v, new_v) in changed_values.items():
                                cli_utils.print_bullet(
                                    key='subkey', value=k, indent=4
                                )
                                cli_utils.print_bullet(
                                    key='old', value=old_v, indent=8
                                )
                                cli_utils.print_bullet(
                                    key='new', value=new_v, indent=8
                                )

                        # print deleted keys
                        deleted_keys = [
                            k for k in old_value if k not in new_value
                        ]
                        if len(deleted_keys) > 0:
                            print('    - deleted subkeys:', deleted_keys)

                    else:
                        cli_utils.print_bullet(
                            key='old value',
                            value=old_config_raw.get(key),
                            indent=4,
                        )
                        cli_utils.print_bullet(
                            key='new value', value=config.get(key), indent=4
                        )

        # make sure file overwrite is confirmed
        if write_new and not overwrite:
            print()
            if not toolcli.input_yes_or_no(
                'Overwrite old config file? ',
                default='yes',
                style=styles['metavar'],
                headless=headless,
            ):
                raise Exception('cannot continue without replacing config file')
    else:
        write_new = True

    print()
    if write_new:
        with open(config_path, 'w') as f:
            json.dump(config, f)
        toolstr.print(
            'Config file created at',
            toolstr.add_style(config_path, styles['description']),
        )
    else:
        print('Config unchanged')




from __future__ import annotations

import toolcli

from ctc import cli
from ctc.protocols import ens_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_records_command,
        'help': 'output text records of ENS name',
        'args': [
            {'name': 'name', 'help': 'ENS name'},
        ],
        'examples': [
            'vitalik.eth',
        ],
    }


async def async_records_command(name: str) -> None:
    text_records = await ens_utils.async_get_text_records(name=name)

    for key, value in sorted(text_records.items()):
        cli.print_bullet(key=key, value=value)



from __future__ import annotations

from ctc import evm
from ctc import spec

from .. import config_values
from . import context_validate


def get_context_chain_id(context: spec.Context) -> spec.ChainId:
    """get chain_id of a given context"""

    context_validate._validate_context(context)
    chain_id, provider = _get_context_chain_id_and_provider(context=context)
    return chain_id


def get_context_network_name(context: spec.Context) -> str:
    """get network_name of a given context"""
    chain_id = get_context_chain_id(context)
    return evm.get_network_name(chain_id)


def get_context_provider(context: spec.Context) -> spec.Provider | None:
    """get provider of a given context"""

    context_validate._validate_context(context)
    chain_id, provider = get_context_chain_id_and_provider(context=context)
    return provider


def get_context_chain_id_and_provider(
    context: spec.Context,
) -> tuple[spec.ChainId, spec.Provider | None]:
    """get chain_id and provider of a given context"""

    from ctc import rpc

    context_validate._validate_context(context)
    chain_id, provider = _get_context_chain_id_and_provider(context=context)

    if provider is not None:
        return chain_id, provider
    else:
        try:
            provider = rpc.find_provider(network=chain_id)
        except LookupError:
            pass
        return chain_id, provider


def _get_context_chain_id_and_provider(
    context: spec.Context,
) -> tuple[spec.ChainId, spec.Provider | None]:

    from ctc import rpc

    if context is None:
        # case: no context provided
        return config_values.get_default_network(), None

    elif isinstance(context, int):
        # case: context is a chain_id
        return context, None

    elif isinstance(context, str):
        # case: context is a network name, provider name, or provider url
        try:
            # try to resolve as network name
            return evm.get_network_chain_id(context), None
        except LookupError:
            # try to resolve as provider
            provider = rpc.resolve_provider(context)
            return provider['network'], provider

    elif isinstance(context, dict):
        # case: context is a dict
        context_network = context.get('network')
        context_provider = context.get('provider')

        if context_provider is not None:
            # subcase: provider is given
            provider = rpc.resolve_provider(context_provider)
            if (
                context_network is not None
                and evm.get_network_chain_id(context_network) != provider['network']
            ):
                raise Exception('context provider does not match network')
            return provider['network'], provider

        elif context_network is not None:
            # subcase: network is given
            return evm.get_network_chain_id(context_network), None

        else:
            # subcase: neither provider nor network are given
            return config_values.get_default_network(), None

    raise Exception('unknown context format: ' + str(type(context)))




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import toolcli

from ctc import spec
from . import trace_crud


async def async_print_transaction_balance_diffs(
    transaction_hash: spec.TransactionHash,
    *,
    normalize: bool = True,
    include_unchanged: bool = False,
    styles: toolcli.StyleTheme | None = None,
    context: spec.Context = None,
) -> None:
    """print balance of a transaction"""
    import toolstr

    if styles is None:
        styles = {}
    toolstr.print_text_box('ETH Balance Diffs', style=styles.get('title'))
    state_diff = await trace_crud.async_get_transaction_state_diff(
        transaction_hash,
        context=context,
    )
    _print_balance_storage_diffs(
        state_diff,
        include_unchanged=include_unchanged,
        normalize=normalize,
        styles=styles,
    )


async def async_print_transaction_storage_diffs(
    transaction_hash: spec.TransactionHash,
    *,
    styles: toolcli.StyleTheme | None = None,
    context: spec.Context = None,
) -> None:
    """print storage diffs of a transaction"""
    import toolstr

    if styles is None:
        styles = {}
    toolstr.print_text_box('Storage Diffs', style=styles['title'])
    state_diff = await trace_crud.async_get_transaction_state_diff(
        transaction_hash,
        context=context,
    )
    _print_storage_diffs(state_diff, styles=styles)


def _print_balance_storage_diffs(
    state_diff: spec.StateDiffTrace,
    *,
    normalize: bool = True,
    include_unchanged: bool = False,
    styles: toolcli.StyleTheme,
) -> None:

    import toolstr

    rows = []
    for address in state_diff.keys():
        row: list[typing.Any] = [address]
        balance = state_diff[address]['balance']
        if isinstance(balance, str) and balance == '=':
            row.extend([None, None, 0])
        elif isinstance(balance, dict):
            if '+' in balance:
                row.extend([0, balance['+'], balance['+']])  # type: ignore
            elif '*' in balance:
                row.extend(
                    [
                        balance['*']['from'],  # type: ignore
                        balance['*']['to'],  # type: ignore
                        balance['*']['to'] - balance['*']['from'],  # type: ignore
                    ]
                )
            elif '-' in balance:
                row.extend([balance['-'], 0, -balance['-']])
            else:
                raise Exception('unknown balance format')
        else:
            raise Exception('unknown balance format')

        if include_unchanged or row[-1] != 0:
            rows.append(row)

    if normalize:
        for row in rows:
            for i in range(1, 4):
                if isinstance(row[i], int):
                    row[i] /= 1e18

    labels = ['address', 'old', 'new', 'diff']
    toolstr.print_table(
        rows,
        labels=labels,
        compact=2,
        label_style=styles['title'],
        border=styles['comment'],
        column_formats={'diff': {'signed': True, 'scientific': False}},
        column_styles={
            'address': styles['metavar'],
            'old': styles['description'],
            'new': styles['description'],
            'diff': styles['content'],
        },
    )


def _print_storage_diffs(
    state_diff: spec.StateDiffTrace, *, styles: toolcli.StyleTheme
) -> None:
    import toolstr

    rows: list[typing.Any] = []
    for contract_address in state_diff.keys():
        contract_address_label = _shrink_hex(contract_address)
        for storage_slot, values in state_diff[contract_address][
            'storage'
        ].items():

            storage_slot_label = _shrink_hex(storage_slot)
            if '+' in values:
                value = values['+']  # type: ignore
            elif '*' in values:
                value = values['*']['to']  # type: ignore
            else:
                raise Exception()

            row = [
                contract_address_label,
                storage_slot_label,
                value,
            ]
            rows.append(row)
            contract_address_label = ''

    labels = ['contract', 'slot', 'new value']
    toolstr.print_table(
        rows,
        labels=labels,
        label_justify='left',
        compact=2,
        border=styles['comment'],
        label_style=styles['title'],
        column_styles={
            'contract': styles['metavar'],
            'slot': styles['description'],
            'new value': styles['content'],
        },
    )


def _shrink_hex(
    hex_str: str,
    *,
    start_chars: int = 6,
    end_chars: int = 4,
) -> str:
    return hex_str[:start_chars] + '...' + hex_str[-end_chars:]

